
"""
    abstract type SiteAmplification

Abstract type to allow dispatch for the `amplification` functions for different site profiles
"""
abstract type SiteAmplification end

"""
    construct_interpolant(fi::T, Ai::T) where {T<:SVector}

Construct a fast interpolant from `StaticArrays` defining the frequencies and amplifications with arbitrary spacing.
Returns a function that takes a frequency input and returns the amplification.
Interpolant is defined to work over the frequency range [0.001, 1000.0) Hz
"""
function construct_interpolant(fi::T, Ai::T) where {T<:SVector}
    f_min = 0.001
    f_max = 1000.0
    num_lnf = 1001
    lnf_min = log(f_min)
    # define initial linear interpolation once using the above irregular grid and extrapolation
    itp = linear_interpolation(fi, Ai, extrapolation_bc=Flat())
    # compute the amp function values at the regular (log-spaced) frequencies
    lnfi = range(lnf_min, stop=log(f_max), length=num_lnf)
    fi_reg = exp.(lnfi)
    Δlnf_reg = lnfi[2] - lnfi[1]
    Ai_reg = itp.(fi_reg)
    # define a new interpolant that is highly efficient and works on the regular-spaced grid
    itp_reg = interpolate(Ai_reg, BSpline(Linear()))
    # return the interpolation function on the regular grid, that takes a scaled logarithmic frequency input
    interpolant = function(f)
        Tf = typeof(f)
        return itp_reg((log(f) - lnf_min) / Δlnf_reg + 1.0)::Tf
    end
    return interpolant
    # return (f) -> itp_reg((log(f) - lnf_min) / Δlnf_reg + 1.0)
end


"""
    struct SiteAmpUnit <: SiteAmplification

Unit amplification function.
"""
struct SiteAmpUnit <: SiteAmplification
    amplification::Function

    function SiteAmpUnit()
        new((f) -> oneunit(typeof(f)))
    end
end

"""
    struct SiteAmpConstant <: SiteAmplification

Constant amplification function
"""
struct SiteAmpConstant <: SiteAmplification
    constant::Real
    amplification::Function

    function SiteAmpConstant(constant::T) where {T<:Real}
        new(constant, (f) -> constant * oneunit(typeof(f)))
    end
end


"""
    struct SiteAmpBoore2016_760 <: SiteAmplification

Implementation of the Boore (2016) amplification function for Vs30 = 760 m/s

# Examples
```julia-repl
	f = 5.0
    sa = SiteAmpBoore2016_760()
	Af = sa.amplification(f)
```
"""
struct SiteAmpBoore2016_760 <: SiteAmplification
    amplification::Function

    function SiteAmpBoore2016_760()
        fi = @SVector [0.001, 0.010, 0.015, 0.021, 0.031, 0.045, 0.065, 0.095, 0.138, 0.200, 0.291, 0.423, 0.615, 0.894, 1.301, 1.892, 2.751, 4.000, 5.817, 8.459, 12.301, 17.889, 26.014, 37.830, 55.012, 80.000, 1e3]
        Ai = @SVector [1.00, 1.00, 1.01, 1.02, 1.02, 1.04, 1.06, 1.09, 1.13, 1.18, 1.25, 1.32, 1.41, 1.51, 1.64, 1.80, 1.99, 2.18, 2.38, 2.56, 2.75, 2.95, 3.17, 3.42, 3.68, 3.96, 3.96]
        new(construct_interpolant(fi, Ai))
    end
end


"""
    struct SiteAmpAlAtikAbrahamson2021_ask14_620 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Abrahamson _et al._ (2014) GMM, for Vs30 = 620 m/s

# Examples
```julia-repl
	f = 5.0
    sa = SiteAmpAlAtikAbrahamson2021_ask14_620()
	Af = sa.amplification(f)
```
"""
struct SiteAmpAlAtikAbrahamson2021_ask14_620 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_ask14_620()
        fi = @SVector [0.001, 0.100000007176552, 0.102329297485949, 0.104712901552804, 0.107151924117949, 0.109647824895946, 0.112201838507376, 0.114815405225881, 0.117489810861085, 0.1202264325802, 0.123026894039802, 0.12589255240632, 0.128824937117037, 0.131825709632546, 0.134896310600769, 0.138038400222065, 0.141253728327524, 0.144543992885956, 0.147910847244603, 0.151356116586059, 0.154881690648375, 0.158489312312177, 0.162181015370784, 0.16595871040885, 0.169824400428869, 0.173780103299513, 0.177828027447085, 0.181970116933159, 0.186208693439478, 0.19054610704757, 0.194984446661483, 0.199526230171352, 0.204173816020742, 0.208929612284355, 0.213796203551149, 0.218776177399838, 0.223872101209225, 0.229086813628305, 0.234422912331458, 0.239883270624677, 0.245470912442472, 0.251188578932815, 0.257039572485661, 0.26302683611382, 0.269153491176368, 0.275422860904314, 0.281838254809002, 0.288403100119285, 0.295120975682464, 0.301995149403153, 0.309029540026393, 0.316227757343325, 0.32359371997251, 0.331131183335223, 0.338844155010958, 0.346736907560462, 0.354813428224786, 0.36307811068037, 0.371535201243506, 0.380189408253386, 0.389045048821185, 0.398107252613262, 0.407380348956526, 0.416869343532483, 0.426579617417572, 0.436515672690002, 0.446683568718976, 0.457088323168438, 0.46773496922304, 0.478630314385223, 0.489778634661086, 0.501187255130982, 0.512861174281571, 0.524807531745957, 0.537031651679684, 0.549540848799636, 0.562341293577962, 0.575440237707926, 0.588843920258061, 0.602559902463521, 0.616594560798692, 0.630957710395976, 0.645654500162755, 0.660693848816064, 0.676082985461142, 0.691831301966495, 0.707945920028427, 0.724435954327771, 0.741310203831773, 0.758578319149634, 0.776247687487606, 0.794327531427445, 0.812831173158537, 0.831763755397804, 0.851137596165377, 0.87096361786093, 0.891250634906882, 0.912011758905137, 0.933255273423412, 0.954993810779107, 0.97723694421474, 0.999999633853153, 1.02329446506163, 1.04713043607716, 1.07151861129162, 1.09647640754566, 1.12201876367331, 1.14815236598752, 1.17489778192726, 1.20226469308904, 1.23026864708192, 1.25892719937347, 1.28824883491205, 1.31825422392517, 1.3489604744997, 1.38038656412012, 1.41253927391579, 1.44543987371868, 1.47911238259783, 1.51355825072426, 1.54881301667269, 1.5848905633235, 1.62181335008947, 1.65958238778041, 1.69824558697017, 1.73779898879862, 1.77828608776837, 1.8197035816424, 1.86208766520357, 1.90546503948543, 1.94984960247986, 1.99525559982877, 2.04173404813673, 2.08928745324258, 2.13795640238707, 2.18776989034643, 2.23872242386584, 2.29087401375753, 2.34423462055453, 2.39884036376148, 2.45470413173396, 2.51189897015488, 2.57038387151322, 2.63027372689445, 2.69152817520359, 2.75424481862134, 2.81838055555125, 2.88404782351028, 2.95120276858626, 3.01997665364079, 3.09027755270178, 3.16230509000893, 3.23591072890429, 3.3113196772554, 3.38845209360351, 3.46738703485683, 3.54810342861765, 3.6307572259458, 3.71538006523731, 3.80192328047219, 3.89049469017383, 3.98103970987384, 4.07377433470771, 4.16865402930154, 4.2658313673159, 4.36514888139538, 4.46688390379777, 4.57086791742096, 4.67741779690484, 4.78635650780478, 4.89775486557788, 5.0118487428658, 5.12859210973947, 5.24809233553927, 5.37040163878342, 5.49543147519744, 5.62335167756751, 5.75435700895086, 5.88845342866043, 6.02563531956834, 6.16588674725113, 6.30967936692519, 6.45651776048592, 6.60689624674967, 6.76082112782712, 6.91828688137004, 7.07959143533954, 7.24441582209855, 7.41306014051683, 7.58587311664298, 7.76249492667347, 7.94328044747604, 8.12822348094524, 8.31774103516475, 8.51139783908034, 8.70962052040683, 8.91240272293129, 9.12024982306876, 9.33264060910502, 9.55009706721427, 9.77261961256125, 10.0001910578874, 10.2327752801061, 10.4710140255772, 10.7149240936345, 10.9645054455084, 11.2205435989428, 11.4814213469585, 11.7487266632786, 12.0224913181046, 12.3027285474753, 12.5894301931157, 12.8825650343572, 13.1820751213882, 13.4902184788619, 13.8035252713909, 14.125541280801, 14.4537496705297, 14.7907638870534, 15.1352457785712, 15.4886663451631, 15.8495022769038, 16.2176073176065, 16.5963627451878, 16.9822974547711, 17.3771048059986, 17.7827957753744, 18.1974302837793, 18.6209015547347, 19.0554232691428, 19.4986547279176, 19.9529560695692, 20.4184012173967, 20.8921832484195, 21.3798822495437, 21.8787484054886, 22.3887129884931, 22.909659227327, 23.4414188072464, 23.9875410094608, 24.4615137952709, 25.0329093218529, 25.6148235017154, 26.2158518759445, 26.8272698465536, 27.4534661797527, 28.09446691795, 28.7502506564272, 29.4207416293659, 30.1117961312001, 30.811505268715, 31.5318949237336, 32.2666451061469, 33.0225752670797, 33.7926308086556, 34.5841820564937, 35.3893524130691, 36.216097853598, 37.0647605768429, 37.9260703333401, 38.8089621672073, 39.7241178868006, 40.6510900911763, 41.599966407565, 42.5707143949687, 43.5632194319038, 44.5772785797191, 45.6125872923706, 46.6833395460697, 47.7757780217904, 48.8893240325792, 50.0400743989436, 51.1942947913361, 52.4039387563387, 53.614957723063, 54.8639426278079, 56.152230540486, 57.4589003466977, 58.8055074891377, 60.168839494608, 61.5723166498931, 63.0170569484297, 64.5041794807385, 66.0052452133238, 67.5480638430476, 69.1334956085096, 70.7283580177624, 72.3998158815981, 74.0787547857259, 75.7996039156837, 77.5626562679972, 79.3680950075974, 81.2610319483109, 83.153414381699, 85.0880264319602, 87.0644638111256, 89.0821208913822, 91.1971203357681, 93.2971759818561, 95.4979057213378, 97.7410868864205, 100.025195173882, 1e3]
        Ai = @SVector [1.0, 1.27131849966856, 1.27364786217161, 1.27603634853953, 1.27848592867056, 1.28099895478186, 1.28357786099471, 1.28622520519563, 1.2889434388998, 1.29173520188997, 1.29460344588835, 1.2975509498314, 1.30058076546606, 1.3036961769696, 1.30690034122135, 1.31019676038717, 1.31358915842919, 1.31708133239743, 1.3206771759199, 1.32438092055858, 1.32819702647961, 1.33212991697744, 1.33618459681898, 1.34036614557702, 1.34467994891009, 1.34913162943125, 1.353727283232, 1.35847298456236, 1.363375527856, 1.36844207057665, 1.37367981612773, 1.3790968414493, 1.3847014608445, 1.39050247305961, 1.39650939948657, 1.40273226553584, 1.40918165695063, 1.41586914366152, 1.42280665323052, 1.43000721599896, 1.43748488902678, 1.44525416858638, 1.45333038870041, 1.46172679282892, 1.47045481180741, 1.47952423032643, 1.48894310672955, 1.49871805295238, 1.50885449107557, 1.51935602406904, 1.5302260715545, 1.54146650237406, 1.55307876107356, 1.56506319254866, 1.57741983233271, 1.5901485779998, 1.60324838288661, 1.61671852084641, 1.63055775591691, 1.64476547137603, 1.65934039368577, 1.67428237920358, 1.68958701478223, 1.70524054060058, 1.72121790294409, 1.73748150584534, 1.75398619883415, 1.77067606186742, 1.78748731066872, 1.80435348984646, 1.82120683669339, 1.83799488215409, 1.85467197895587, 1.87120483603363, 1.88756584928089, 1.90373730740876, 1.91970625040889, 1.93546650295583, 1.9510152593158, 1.96635530570864, 1.98149114618527, 1.99643229005328, 2.01118329485831, 2.02575247931318, 2.04014641122666, 2.05437358672245, 2.06844032117375, 2.08235468419301, 2.09612436216508, 2.10975761252428, 2.12326084225104, 2.13664192539445, 2.14991154487293, 2.16307443060873, 2.17614033344739, 2.18911787206594, 2.20201423012552, 2.21483920620629, 2.22759955370566, 2.24030456464251, 2.25296154148033, 2.26557971725804, 2.27816467473245, 2.29071788271264, 2.3032400469491, 2.31572599647585, 2.32815556505631, 2.34054940209122, 2.35290715713082, 2.36522238277542, 2.37749051831689, 2.38970715274537, 2.4018633734267, 2.41395490786702, 2.42597554460696, 2.43791940595674, 2.4497757462845, 2.4615391505608, 2.47320480678957, 2.4847601322214, 2.49620408246273, 2.50752668443819, 2.51871060672102, 2.5297596074038, 2.54067584282881, 2.55144664044525, 2.56207261663909, 2.57254204204221, 2.58285360931613, 2.59300361000181, 2.60298551195721, 2.61279323256054, 2.6224286470533, 2.63188354354167, 2.64115762996102, 2.65024527504853, 2.65912810778427, 2.66782060123297, 2.67631755269486, 2.68461815999178, 2.69271817986727, 2.70062213382974, 2.70831871725709, 2.71581770156889, 2.72310883251496, 2.73019894373201, 2.73707865846449, 2.74375568118252, 2.75022140641828, 2.75648433771688, 2.76253232665129, 2.76837869241064, 2.77400770979689, 2.77943316921774, 2.7846347466576, 2.78961027235924, 2.79437303506795, 2.79892953960913, 2.80327862430555, 2.80741548398516, 2.81134322510281, 2.81505764409859, 2.81856591128218, 2.82186444735045, 2.82495695789317, 2.82783721427689, 2.83051212126031, 2.83297633132674, 2.83523639953662, 2.8372877156902, 2.83913143401513, 2.84077094014203, 2.84220512105643, 2.84338239396761, 2.84453752539754, 2.84569170133826, 2.84684594137068, 2.84800140850567, 2.84915737319923, 2.85031344233642, 2.85146887884798, 2.85262696721866, 2.85378303053006, 2.85494041625843, 2.85609851104183, 2.85725669775932, 2.85841651300168, 2.85957503865203, 2.86073384714995, 2.86189466638097, 2.86305449974859, 2.86421506638288, 2.86537569117408, 2.86653838525032, 2.86769978356148, 2.86886195554912, 2.87002416833799, 2.87118870499014, 2.87235206291314, 2.87351642567287, 2.87468120313107, 2.87584574300883, 2.87700920982724, 2.87817424036366, 2.87934030037972, 2.880506679293, 2.88167640401134, 2.88284155564091, 2.88400866295102, 2.88517716959085, 2.88634646120449, 2.88751592098892, 2.88868480884237, 2.88985237467466, 2.89102661665717, 2.89219376110607, 2.89336648561349, 2.89453497549384, 2.895707904823, 2.89687997662864, 2.89805545284443, 2.89922873972961, 2.90039879683353, 2.90157572463167, 2.90274806501943, 2.90392049668863, 2.90509824676344, 2.90627496465686, 2.90744978512103, 2.90862824445576, 2.90980342588609, 2.91098091930798, 2.91216025898202, 2.91333379520812, 2.91451474592521, 2.91569566571787, 2.91687578300107, 2.91805431552053, 2.91923034618551, 2.92041104624385, 2.92141282611799, 2.92259748894795, 2.92377684076637, 2.92496752931683, 2.92615159941094, 2.92733703087574, 2.92852325524016, 2.92970958073521, 2.93089524852177, 2.93208979416362, 2.93327213905911, 2.93446206595989, 2.93564847728727, 2.93684161114483, 2.93802978805933, 2.93922366236317, 2.94041080702049, 2.94160236434546, 2.94279806654809, 2.94398427376105, 2.94517293468467, 2.94637732449283, 2.94756988100288, 2.94876318656724, 2.94995654384797, 2.95114925098669, 2.95234047346702, 2.95352936452941, 2.9547313446324, 2.9559301177744, 2.95712457557147, 2.95833123099914, 2.95951433263599, 2.96072642625698, 2.96191259644968, 2.96310852156839, 2.96431434309621, 2.96550991642039, 2.96671434950427, 2.96790632397305, 2.96910584323222, 2.97031288839944, 2.97152742012084, 2.9727258335987, 2.97392991530171, 2.97513951470086, 2.97632893885952, 2.97754748893125, 2.97874405828706, 2.97994290437603, 2.98114356152975, 2.98234550965926, 2.98357737374197, 2.98478124669835, 2.98598439797552, 2.98718591838467, 2.98838495124345, 2.98961366346154, 2.99080627211428, 2.9920279941566, 2.99324532515824, 2.99445681071514, 2.99445681071514]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_ask14_760 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Abrahamson _et al._ (2014) GMM, for Vs30 = 760 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_ask14_760 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_ask14_760()
        fi = @SVector [0.001, 0.102329291629789, 0.104712890010164, 0.107151914957207, 0.109647813102497, 0.112201836659156, 0.114815393694835, 0.117489795185623, 0.120226421637391, 0.123026889157653, 0.125892536389367, 0.128824935296206, 0.13182570291595, 0.134896294354494, 0.138038411921428, 0.141253715584794, 0.144543993245758, 0.147910831725714, 0.151356107872253, 0.154881694399225, 0.158489303339742, 0.162180978112495, 0.165958692627976, 0.169824387200124, 0.173780087479412, 0.177827987159618, 0.181970098272324, 0.186208709701579, 0.1905460700907, 0.194984407472064, 0.199526184904639, 0.204173763555912, 0.208929612563225, 0.213796165486253, 0.218776181039101, 0.223872120331069, 0.229086778100748, 0.234422848322377, 0.239883277899045, 0.245470899315408, 0.251188601619759, 0.257039480147694, 0.263026791244529, 0.269153392326978, 0.275422840939865, 0.281838278374961, 0.288403066962298, 0.295120927437312, 0.301995171485957, 0.309029453737916, 0.316227724229087, 0.323593559416164, 0.331131050939781, 0.338844123996532, 0.346736740907403, 0.354813422978372, 0.363078118172697, 0.37153517036869, 0.380189278371953, 0.389045079100575, 0.398107032945674, 0.407380141046479, 0.416869301624597, 0.426579390461964, 0.436515671631233, 0.446683430400802, 0.457088039005123, 0.46773511556702, 0.478629897807854, 0.489778872693707, 0.50118719863843, 0.512861176181064, 0.524807303576546, 0.537031671638095, 0.549540677985814, 0.562341420837777, 0.575439915771869, 0.588843687408805, 0.602559095335443, 0.616594498178253, 0.630956838282629, 0.645654416042027, 0.660693435579826, 0.676082666603497, 0.691830720734741, 0.707945795957969, 0.72443532781887, 0.741309624084325, 0.758577203919982, 0.77624745769222, 0.794328291514734, 0.812830676987017, 0.831763916668987, 0.851137859274074, 0.870962885444047, 0.891249875528529, 0.912010167053346, 0.933253679988854, 0.954992166274483, 0.977237551755246, 0.999999168813753, 1.0232925177164, 1.04712953459331, 1.07151857211678, 1.09647732276226, 1.12201805278287, 1.14815183764178, 1.17489782100591, 1.20226477971155, 1.23026962056445, 1.25892432815131, 1.28825074454752, 1.31825825555489, 1.34896263627262, 1.38038268450569, 1.41253552190051, 1.44544086549261, 1.47910458530454, 1.51356159519225, 1.54881530453188, 1.58489389973832, 1.62180870243341, 1.65958562234269, 1.69824588904663, 1.73779618369895, 1.77827624579108, 1.8196950734287, 1.86208930506452, 1.90545964702916, 1.94984645250691, 1.99526156164151, 2.04173991259483, 2.08929510892326, 2.137953540857, 2.18775791681998, 2.23871166806337, 2.29086258407858, 2.3442317459986, 2.39882309610215, 2.45471229794924, 2.51188920081331, 2.57039958766811, 2.63027351343954, 2.69154338255738, 2.7542199774598, 2.8183904204513, 2.88401933081082, 2.95120421062146, 3.01993722938914, 3.0903013725065, 3.1622906167405, 3.23592899478649, 3.31131792994546, 3.38845538724139, 3.46737641623156, 3.54811898133175, 3.63077222275196, 3.7153395206109, 3.80187246745282, 3.89042877823982, 3.98107343300018, 4.0738177070868, 4.16866925869073, 4.26577105198063, 4.36514527469599, 4.46681233698773, 4.57087241800643, 4.67735553529616, 4.7862938382588, 4.89782051909636, 5.01186817841011, 5.12857475720548, 5.24810170023344, 5.37028292809851, 5.49540706428733, 5.6234241270191, 5.75440312419696, 5.8884201663215, 6.02555999282607, 6.16591694805352, 6.3095969885328, 6.45646411719527, 6.60690986461733, 6.7607663559943, 6.91819515641402, 7.07936445998824, 7.2442336646087, 7.4129855022168, 7.58582915560075, 7.76246351980547, 7.94336389185347, 8.12819629982869, 8.31748830322892, 8.51120311935212, 8.70963007766765, 8.91236851110203, 9.12007863057234, 9.3323253763062, 9.54985042066809, 9.77218486971381, 10.0001661852962, 10.2327950905944, 10.4714698810819, 10.7151053301551, 10.9646942099711, 11.2202000511668, 11.481556688206, 11.7486642307561, 12.0228136506425, 12.3025402876746, 12.5892110270895, 12.882800311789, 13.1823742451315, 13.489549091985, 13.8033527682673, 14.1256461792077, 14.4543590459429, 14.7914832545465, 15.1358306960742, 15.4884466073567, 15.849255728102, 16.218139360028, 16.5963896654057, 16.9824668860576, 17.3777129943964, 17.7820306844788, 18.197062425104, 18.6209884397816, 19.0535517185925, 19.4986028224736, 19.9519735653357, 20.4178297858679, 20.8938834506942, 21.3798774703834, 21.8781767449478, 22.3859358255223, 22.9086386304349, 23.4434555840424, 23.9868906474195, 24.4617728210289, 25.030627194913, 25.6181897205504, 26.2130148419347, 26.8266427207605, 27.455172621907, 28.0937811170387, 28.751449890638, 29.4234015101268, 30.1092477949067, 30.8142629304006, 31.5326307205745, 32.2700319384226, 33.019847133834, 33.7952661610636, 34.5823741527521, 35.3877712008252, 36.2111912675171, 37.0609536812008, 37.9287301761387, 38.8139433054402, 39.7158673476879, 40.6442831286205, 41.5997404939946, 42.5709162935017, 43.5564642909024, 44.5810126496729, 45.6191146627162, 46.6833433167586, 47.7736556953069, 48.8898915488733, 50.0317617900903, 51.1988284390013, 52.3904910767017, 53.6257146986582, 54.86494081583, 56.1476929120039, 57.4528069787863, 58.8029351241702, 60.1743686855608, 61.5919359491351, 63.0290608028603, 64.4831587917317, 65.9825678305874, 67.5286047970775, 69.1225735455047, 70.7292127373549, 72.3826406821742, 74.0837712670898, 75.7908268990384, 77.5877527784319, 79.3872762217308, 81.2324006210061, 83.1231186036943, 85.0592352916312, 87.0403434241269, 89.0657885682727, 91.1988372294617, 93.31288331588, 95.4676758375373, 97.736133375685, 99.9701348747591, 1e3]
        Ai = @SVector [1.0, 1.06703864090064, 1.06875586797083, 1.07052163674707, 1.07233767092153, 1.07420559831727, 1.07612720782859, 1.07810426753169, 1.08013864965441, 1.08223245944135, 1.08438765651816, 1.08660643976537, 1.08889111257429, 1.09124392886396, 1.0936674099055, 1.09616412964452, 1.09873685225255, 1.10138828109655, 1.10412144985124, 1.10693950131562, 1.10984556771971, 1.1128431741543, 1.11593592797726, 1.11912755522829, 1.12242200847979, 1.12582355013723, 1.12933647239317, 1.13296549688731, 1.13671552415173, 1.14059166416224, 1.14459947837975, 1.14874470649292, 1.15303347155909, 1.15747217527575, 1.16206785456776, 1.16682764058575, 1.1717593713337, 1.17687122452897, 1.18217209241337, 1.18767127364443, 1.1933786811155, 1.19930504685837, 1.20546194467564, 1.21186127452283, 1.2185164725599, 1.22544143262654, 1.2326512596971, 1.24016254091402, 1.24799260877242, 1.25616050094805, 1.26468707947706, 1.2735944077208, 1.2829069921853, 1.29265118602082, 1.30285389938649, 1.31343991498885, 1.32435199509725, 1.33559513989266, 1.34716756579281, 1.35906051889546, 1.37125849333225, 1.38374060422001, 1.39648017467787, 1.4094453018432, 1.42259992584518, 1.43590386337062, 1.4493134964406, 1.46278257976503, 1.4762620608898, 1.48970313149132, 1.50305925969573, 1.51629469007934, 1.52938261848903, 1.54230348253616, 1.55504478055455, 1.56759746094461, 1.57995140591516, 1.59212079860389, 1.60411040817598, 1.61592672351486, 1.62757481023298, 1.63906073226851, 1.65038876459038, 1.66156504265066, 1.67259543567938, 1.68348536624952, 1.69423966460146, 1.70486495800563, 1.71536659731093, 1.72575036157773, 1.73602109634979, 1.74618529536434, 1.75624844105561, 1.76621620848404, 1.77609444960205, 1.78588917663348, 1.79560652952295, 1.80525187170047, 1.81483135608769, 1.82435110438271, 1.83380601841381, 1.84321060533599, 1.85257372146419, 1.86189989720152, 1.87119722716341, 1.88047164444702, 1.88972866173761, 1.89897657283838, 1.90822005722035, 1.91746655581289, 1.92672185454918, 1.93599454630268, 1.94528802955623, 1.95460611351215, 1.96395205545783, 1.97332730096478, 1.98273292143254, 1.99216504568077, 2.00162710949167, 2.01111304111233, 2.02062288684835, 2.03015160820196, 2.03969749825877, 2.04925721106831, 2.05882351678941, 2.06839692816824, 2.07797039280434, 2.08754308076974, 2.09710589275194, 2.1066585347477, 2.11619441379135, 2.12571175557418, 2.13520445916782, 2.1446690118278, 2.15410507762589, 2.16350489052618, 2.17286898918743, 2.18219307462439, 2.19147010421682, 2.20070531634737, 2.20988960997688, 2.21902327005749, 2.22810402779576, 2.23713014542967, 2.24609682660765, 2.25501024943307, 2.26385952749534, 2.27265206044218, 2.2813812360854, 2.29005222814591, 2.29865916703363, 2.30720011746131, 2.31568211747521, 2.32410039642848, 2.33245450833281, 2.34074443847271, 2.34897537825235, 2.35714380777201, 2.36525112972899, 2.37329932311726, 2.38129096270607, 2.38922384784986, 2.39709585164769, 2.40491607654021, 2.41268355383092, 2.42039742245455, 2.42806297630066, 2.4356801397874, 2.4432482878922, 2.45077192942042, 2.45824908637873, 2.46568721066236, 2.47309486262213, 2.48046030626912, 2.48779963178117, 2.49510858410167, 2.50239001812315, 2.50964719361091, 2.51688373994356, 2.52410380233782, 2.53131197866005, 2.53847199813543, 2.54567344943883, 2.55289148164837, 2.56012610679884, 2.56738118224535, 2.57465113016471, 2.58194018564737, 2.58925340653725, 2.59657426728633, 2.60391872734924, 2.6112694708612, 2.61864366195085, 2.62603595781132, 2.63345331423361, 2.64087701333467, 2.64832730374271, 2.65578474655584, 2.66327146067633, 2.67076739569765, 2.67829669322392, 2.6858224787706, 2.6933861205713, 2.7009492293435, 2.70853884164594, 2.71614966449537, 2.72377572648241, 2.73141037067914, 2.73908620702659, 2.74675827002529, 2.75446014579853, 2.76218683742717, 2.7699100106748, 2.77766734440365, 2.7854302232962, 2.79324032632552, 2.80104324587815, 2.80888175823791, 2.81672477235592, 2.82459207194985, 2.83247784701687, 2.84037545866793, 2.84830823844775, 2.8562398498425, 2.86419400677171, 2.87216448824192, 2.88017916580779, 2.88819854391081, 2.89621418090161, 2.90429286386946, 2.91235462550598, 2.92046936044239, 2.92859258203706, 2.9367161877715, 2.94487552787013, 2.95301996803807, 2.96123304387097, 2.96946515777201, 2.97765917855897, 2.98467401591372, 2.99293053573537, 3.00128425508161, 3.00956834985056, 3.01793944199022, 3.02633874110576, 3.03469831060109, 3.04313136509235, 3.05157156429415, 3.06001024454111, 3.0685075676036, 3.07698897569764, 3.08551723420504, 3.0940118092181, 3.10261703281621, 3.11117372728503, 3.11975023387862, 3.12833938497636, 3.1370223080821, 3.14570846940482, 3.15438835859857, 3.16305136839715, 3.17178669857566, 3.18059320558372, 3.18936213624136, 3.19807901730661, 3.20695586331613, 3.21576683056899, 3.22461519007999, 3.23349546791687, 3.24240152116193, 3.2513263387366, 3.26026211950977, 3.26920006983149, 3.27827611014175, 3.28719598262357, 3.29624070601799, 3.30525544156978, 3.31439112122328, 3.32348222079511, 3.33268797330732, 3.34183110712135, 3.35089374047133, 3.36004796521199, 3.36929457925546, 3.37863434820875, 3.38785704285932, 3.39715515061964, 3.40652726934536, 3.41574060072069, 3.42524186793735, 3.434563876053, 3.44392739058654, 3.45332707129066, 3.46275674855456, 3.4722093556338, 3.48167703081897, 3.49144617416317, 3.50093264847628, 3.51040514481747, 3.5201753622639, 3.52959082307405, 3.52959082307405]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_ask14_1100 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Abrahamson _et al._ (2014) GMM, for Vs30 = 1100 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_ask14_1100 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_ask14_1100()
        fi = @SVector [0.001, 0.0977009871629398, 0.0999999974066191, 0.102329292723223, 0.10471289115513, 0.107151916156133, 0.109647814357927, 0.112201837973753, 0.114815395071387, 0.117489796627051, 0.12022642314675, 0.123026890738146, 0.125892538044346, 0.128824937029181, 0.131825698037974, 0.134896294629028, 0.138038396066939, 0.141253712785238, 0.144543994114778, 0.147910821002263, 0.151356111737264, 0.15488169624427, 0.158489284937044, 0.162180990779373, 0.165958686142594, 0.169824397487713, 0.173780077682896, 0.177827989905068, 0.181970075789458, 0.186208696802254, 0.190546089530823, 0.194984390967264, 0.19952619872714, 0.204173791156713, 0.208929578476893, 0.213796157219783, 0.218776182483442, 0.223872073974258, 0.229086805382967, 0.234422887101731, 0.239883242654684, 0.245470849447296, 0.251188575845802, 0.257039494058099, 0.263026750987176, 0.269153396051622, 0.275422838647827, 0.281838211787, 0.288403047254166, 0.295120911515252, 0.30199515230416, 0.309029461082775, 0.316227810959307, 0.323593528220054, 0.331131014939867, 0.338843995274008, 0.346736787224427, 0.354813408075398, 0.363078003101169, 0.371535229845431, 0.380189370412531, 0.389045015362505, 0.398107129653939, 0.407380169972867, 0.416869387998658, 0.426579460490376, 0.436515642062981, 0.446683603942946, 0.457088011726756, 0.467735030372381, 0.478630059436938, 0.48977886174253, 0.501186900923732, 0.512861101135527, 0.524807426977635, 0.537031806989445, 0.549540539495192, 0.562341256619687, 0.575439855356368, 0.588843529560242, 0.602559464658751, 0.616594793588158, 0.630957122685767, 0.645653967098973, 0.660693346609629, 0.676082476533705, 0.691830380094231, 0.707945190893355, 0.724435846540562, 0.741309834383654, 0.758576929235547, 0.776246740734158, 0.794327667258409, 0.812830478038647, 0.831763668856544, 0.851138257025319, 0.870963800915754, 0.891250342744557, 0.912011002228203, 0.933253426960802, 0.954992169704093, 0.977237337184143, 0.999999431520578, 1.02329258802432, 1.04712873284538, 1.07151870290434, 1.09647737309909, 1.12201694252587, 1.14815218303086, 1.17489685409032, 1.20226345909574, 1.23026747271141, 1.25892554890496, 1.28824801873988, 1.31825585970448, 1.34896088340374, 1.38038380739445, 1.41253790846537, 1.44543991146322, 1.47910741518123, 1.51355897881738, 1.54881424565928, 1.58489013888922, 1.62180826292029, 1.65958737746723, 1.69824256006834, 1.73779828792432, 1.7782759683263, 1.81969809878839, 1.86208300469775, 1.90546073274021, 1.94984206152573, 1.99525798039047, 2.04173612727008, 2.08929487780124, 2.13796268150687, 2.18775720016312, 2.23872156828875, 2.29086319409698, 2.34422279927247, 2.39882720656895, 2.45470437951388, 2.5118831262352, 2.57039286287139, 2.63026340393243, 2.69153643345486, 2.75423092461459, 2.81837601483687, 2.88402755531517, 2.95120221165731, 3.01994301176439, 3.0902943436397, 3.16226791918128, 3.23592343978858, 3.31130639900788, 3.38844346520425, 3.46735829075066, 3.54813796982173, 3.63076229148808, 3.71534669081613, 3.801892180906, 3.89044610656452, 3.98105751405223, 4.07380854238369, 4.16869221231179, 4.26579468976477, 4.36513953987965, 4.46682411265725, 4.57087674263924, 4.67732286510221, 4.78627663067936, 4.89777211817101, 5.01184183117266, 5.12862521542859, 5.24805454537812, 5.37028060850859, 5.49541580776269, 5.62338453092948, 5.75437227828415, 5.88844607464468, 6.02559802172859, 6.16589398182084, 6.30958607432139, 6.45649876349053, 6.60690890495408, 6.760828758235, 6.91826280879999, 7.0794443841581, 7.24439895822411, 7.41301413135051, 7.58570315772501, 7.7623709696718, 7.94320167603534, 8.12824505786064, 8.31755037303491, 8.51135148759121, 8.70953361675237, 8.91256121414362, 9.12012223126881, 9.33243620666656, 9.54994255437179, 9.77238579015416, 9.99998079300324, 10.232879861281, 10.4712538288613, 10.7152949734939, 10.9648844956418, 11.2202146864893, 11.4815052803655, 11.7490074721504, 12.0225917857968, 12.3025220369835, 12.589103564177, 12.8821971342623, 13.1826526141445, 13.4892888965139, 13.8035425565483, 14.1253154493635, 14.4544590583405, 14.7907677635153, 15.1354174421594, 15.4882980859825, 15.8484371645514, 16.2180328252238, 16.5961517672835, 16.9825665808237, 17.3779810898009, 17.7821741443577, 18.1970908347437, 18.6203446881932, 19.053982058617, 19.4978468737616, 19.9531057948348, 20.4167255747922, 20.8927512271844, 21.3795343032608, 21.8767465388772, 22.3876321461927, 22.9083678904629, 23.4424212027064, 23.987524339405, 24.4611328388709, 25.0316992744836, 25.6170945647467, 26.2146717990403, 26.8266726818231, 27.4528300428137, 28.0927699486552, 28.7493251169649, 29.4223987334913, 30.108087751031, 30.813343104761, 31.5300579809116, 32.2657695904908, 33.0204385313945, 33.7939274806613, 34.5808256137391, 35.3908395432242, 36.212760526797, 37.0632487171366, 37.9240601780719, 38.8135353154471, 39.7182173175756, 40.6442660545585, 41.5912069613362, 42.5668541119602, 43.5627905245868, 44.5780338296613, 45.6213470331911, 46.6823109274263, 47.770375329313, 48.885319197145, 50.026760373473, 51.1941325267533, 52.4005309523462, 53.617862354211, 54.873591949379, 56.1522828539384, 57.469459825696, 58.807519879041, 60.1833856512864, 61.5769801307731, 63.0282396696194, 64.4949557174051, 65.9973082892216, 67.534697485468, 69.1064003046458, 70.7394762496497, 72.3771421349366, 74.0762326373562, 75.8061756762341, 77.5645386722037, 79.3856821717277, 81.2327436035212, 83.1435705160823, 85.0764783644605, 87.0731307538662, 89.0863780489238, 91.1620407295812, 93.3015037249044, 95.5060564600769, 97.7157245280178, 99.986853832442, 1e3]
        Ai = @SVector [1.0, 1.07272009904468, 1.07461824650246, 1.07655171146224, 1.07854109520397, 1.08058819823652, 1.0826951497798, 1.08486397812369, 1.08709690939917, 1.08939615726736, 1.09176406893817, 1.09420327652032, 1.09671625725455, 1.09930578290889, 1.10191017953457, 1.10452493872894, 1.10715174175041, 1.10979243811057, 1.11244890835546, 1.11512289913659, 1.11781643572506, 1.12053128512129, 1.12326817918572, 1.12602703376797, 1.12880658505047, 1.13160477706513, 1.13441869676906, 1.13724495190875, 1.14007928263506, 1.14291720161923, 1.14575415957275, 1.14858592660218, 1.1514089643221, 1.15421987995924, 1.15701571909552, 1.15979392670495, 1.16255225966834, 1.1652885983998, 1.16800137297815, 1.1706889570505, 1.1733501619506, 1.1759840652315, 1.17858995753545, 1.18116744880569, 1.1837163884907, 1.18623676205161, 1.1887288695556, 1.19119303022579, 1.19362984302879, 1.19604001630278, 1.19842427160126, 1.20078352793097, 1.20311886539246, 1.20543119825006, 1.20772183720017, 1.20999191388274, 1.21224273635681, 1.21447550065783, 1.21669141739463, 1.21889180892343, 1.22107784804006, 1.22325078247076, 1.22541190626654, 1.22756238349845, 1.2297035270306, 1.23183650163022, 1.23396258564557, 1.23608311850285, 1.23819921920532, 1.24031229009923, 1.24242356583984, 1.24453433506328, 1.24664581247653, 1.24875948751423, 1.25087663538713, 1.25299853195304, 1.25512650122771, 1.25726208863782, 1.25940652727078, 1.26156125549539, 1.26372771341754, 1.26590732199988, 1.26810155720129, 1.27031189425135, 1.27253985810081, 1.27478687788863, 1.27705462288155, 1.27934276014941, 1.28165462411655, 1.28399198654355, 1.28635664067962, 1.28875037455688, 1.29117481695682, 1.29363193024651, 1.29612338951354, 1.29865122116436, 1.30121727928775, 1.30382351199079, 1.30647227682148, 1.30916528339693, 1.31190513928398, 1.31469392835841, 1.31753382854405, 1.32042751764099, 1.3233774564062, 1.32638600650979, 1.32945610373235, 1.3325904068269, 1.3357919662761, 1.33906372372302, 1.34240837939107, 1.34582891467241, 1.34932837994988, 1.35290898209872, 1.35657421353127, 1.36032641355398, 1.36416901384907, 1.36810455249425, 1.37213602407855, 1.37626657040445, 1.38049951606509, 1.38483839009469, 1.38928649325937, 1.39384780778565, 1.39852608936178, 1.40332475364739, 1.40824856544542, 1.41330205729258, 1.41849009964461, 1.42381723127328, 1.42928970963458, 1.43491029213688, 1.44068661703736, 1.44662599002522, 1.45273328458737, 1.45901308812765, 1.46546705130721, 1.47209871424575, 1.47890568795647, 1.48588873879748, 1.4930457675397, 1.50037382724259, 1.50786922046871, 1.5155275540678, 1.52334376876346, 1.53131376413294, 1.53942995149464, 1.54768560954972, 1.55607705944115, 1.56459529185976, 1.57323440509721, 1.58198845237596, 1.59084726340846, 1.59980654657903, 1.60886015557629, 1.61799967868031, 1.6272164217661, 1.63650914557563, 1.64586429703662, 1.65528391515161, 1.66475725330086, 1.67427892813077, 1.68384389875844, 1.69345075089038, 1.70308898687821, 1.71275793566168, 1.72245078705748, 1.73216819740151, 1.7419043416768, 1.7516534515231, 1.76141811934267, 1.77119384316637, 1.78097628509195, 1.79077051451167, 1.80056435016061, 1.81036399461194, 1.8201724868708, 1.82997835699894, 1.83979059989863, 1.84960922116068, 1.85942898812654, 1.86925038215246, 1.87908693190678, 1.88892282798721, 1.89877320923897, 1.90863559957257, 1.91850741363785, 1.92840065402361, 1.93831442165585, 1.94823997845153, 1.95819986775019, 1.96818692768151, 1.9782102887716, 1.98827162803559, 1.99837290811246, 2.00852605664248, 2.01872488554179, 2.02899325647068, 2.03931573558146, 2.04967418993201, 2.0601133763512, 2.07060372382565, 2.08113528731286, 2.09170972954, 2.10232933167292, 2.11299710662163, 2.12370244067942, 2.13444816749834, 2.14523802190605, 2.15607680791983, 2.1669537059346, 2.17787370041792, 2.1888430487277, 2.19985080665347, 2.21092307478773, 2.22201077493362, 2.23316030693854, 2.24436218544429, 2.2556054542483, 2.26687757989132, 2.27821218094849, 2.28959953097319, 2.3010027281741, 2.31248545169504, 2.32401250633251, 2.33557124555343, 2.34717697245005, 2.35881738764232, 2.37054226352417, 2.3822781308288, 2.39407600270815, 2.4059253507225, 2.4178508849077, 2.42976745923843, 2.44177306889645, 2.45381964595755, 2.46589329495152, 2.47806620561788, 2.49024109523032, 2.50249301689143, 2.51476373848892, 2.52522141545063, 2.53761560316165, 2.55010070088297, 2.56260604499172, 2.57517234514276, 2.58778760568597, 2.60043804807758, 2.61317290555668, 2.62598316336768, 2.63878838235406, 2.65171176644399, 2.6645984212046, 2.67757777859987, 2.69064164695608, 2.70378040744139, 2.71689614902714, 2.73014380516897, 2.74333383153553, 2.75672624835215, 2.77002707680706, 2.78351299093379, 2.79697254163969, 2.81049145410968, 2.82405589121605, 2.83776993956185, 2.85150757506944, 2.86524932646822, 2.87910670960228, 2.89293465834964, 2.90684990634723, 2.92084194148866, 2.9348985163986, 2.94900553381421, 2.96331261610237, 2.97748015891666, 2.99182138611405, 3.00615236768523, 3.02063939169015, 3.03508181862186, 3.04965531487661, 3.06414077639189, 3.07894459447328, 3.09362811047786, 3.10838779315418, 3.12321003239785, 3.13807573868171, 3.15323511660044, 3.16815460160513, 3.18334535102935, 3.19852502461294, 3.21366680703337, 3.22905782659867, 3.24437839387357, 3.25993439598707, 3.27537896464421, 3.29103764303801, 3.30653383328533, 3.32221368542922, 3.33807718010097, 3.35412348943361, 3.36991091436369, 3.3858217188871, 3.3858217188871]
        new(construct_interpolant(fi, Ai))
    end
end


"""
    struct SiteAmpAlAtikAbrahamson2021_bssa14_620 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Boore _et al._ (2014) GMM, for Vs30 = 620 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_bssa14_620 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_bssa14_620()
        fi = @SVector [0.001, 0.0988597583233483, 0.10000000905698, 0.102329313938003, 0.104712907225502, 0.107151928415835, 0.109647833806917, 0.112201853536817, 0.114815416755257, 0.11748980899071, 0.120226447690601, 0.123026907889884, 0.125892549490764, 0.128824947136627, 0.131825713083256, 0.134896318640697, 0.138038418217646, 0.141253742621104, 0.144544010331342, 0.147910858264061, 0.151356135757439, 0.154881715984271, 0.158489343800596, 0.162181024977386, 0.165958714245848, 0.169824426384122, 0.173780119618491, 0.177828026653765, 0.181970149492778, 0.186208745413493, 0.190546136191941, 0.194984472771739, 0.199526244867303, 0.204173849648116, 0.208929632051863, 0.213796259156406, 0.218776225790489, 0.223872160502327, 0.229086863604599, 0.234422963882157, 0.239883331593177, 0.245470952735973, 0.251188631268231, 0.257039566972664, 0.263026867052165, 0.269153479193755, 0.275422981993268, 0.281838430665289, 0.288403183240027, 0.295121036430249, 0.30199532016941, 0.309029681385659, 0.316227959453922, 0.323593815430701, 0.331131125293165, 0.33884413965573, 0.346736844105946, 0.354813571718636, 0.363078097884058, 0.37153533489662, 0.380189432953379, 0.389045387684826, 0.398107382434792, 0.40738045458149, 0.416869559055489, 0.426579848570301, 0.43651596623884, 0.44668400463115, 0.457088385344204, 0.467735261703718, 0.478630157887024, 0.489778747086552, 0.501187335766022, 0.512861581142204, 0.524807498332346, 0.537032210612533, 0.549540852571808, 0.562341870630425, 0.57544046521564, 0.588844308449063, 0.602559995333324, 0.616595844987953, 0.630957558958681, 0.645654794797108, 0.660694177542891, 0.676083899673177, 0.691830990537161, 0.707945672573651, 0.724437042167066, 0.741310651761912, 0.758578903635447, 0.776248662527666, 0.794328058553149, 0.812831315341774, 0.831763429555492, 0.85113774549701, 0.870964759497601, 0.89125283141736, 0.912012782125235, 0.933254304756467, 0.954991789969289, 0.977238521026131, 1.0000023432413, 1.02329199301602, 1.04713005392984, 1.07152077887013, 1.09647780112639, 1.12202176953535, 1.14815566129574, 1.17489868642837, 1.20226720726246, 1.23026752889531, 1.25892566850813, 1.28825121936761, 1.31825790247562, 1.34896343274778, 1.38038719333504, 1.41253637652755, 1.44544512052703, 1.47910568445794, 1.51356696971057, 1.54882323535081, 1.58489888950969, 1.62180984611921, 1.65958587061972, 1.69824796089502, 1.73780552331717, 1.77828217420365, 1.81970422254956, 1.86208515275049, 1.9054724229932, 1.94985012740896, 1.99527070787577, 2.04173630344276, 2.0893083951811, 2.13797175768858, 2.18777328777465, 2.23871864068967, 2.29087073444947, 2.34423024579319, 2.39883177562053, 2.45471482650006, 2.51189259796148, 2.5704096362536, 2.63028194393451, 2.69152489765904, 2.75422829217587, 2.81837812414992, 2.88403590342629, 2.95122773545922, 3.01998066568673, 3.09032260868609, 3.16228244386608, 3.23594330407204, 3.31134484354241, 3.38847095643201, 3.46735834163272, 3.54817498394486, 3.63078349322634, 3.71536572639437, 3.80190149002748, 3.8904465796713, 3.98106213681025, 4.07381551219341, 4.16869054376005, 4.2658511826137, 4.36518885955019, 4.46687936557587, 4.57090821709747, 4.67736536907894, 4.78635192086503, 4.89785462271124, 5.01184961533696, 5.12857918519133, 5.24806034044961, 5.3703242692595, 5.49549088729441, 5.62338735195922, 5.75448188593756, 5.88839072224789, 6.0256295160731, 6.16597957071741, 6.30960270593151, 6.45668573120628, 6.60695695440333, 6.76084384119688, 6.91831606724392, 7.07960762923677, 7.24439292930821, 7.41320680717793, 7.58569585774497, 7.76246077056568, 7.94348611541068, 8.12835237807942, 8.31775252136749, 8.51123657664752, 8.70959011881566, 8.91280701501557, 9.12036374628954, 9.33265604127413, 9.55015246249384, 9.77224591631989, 10.0000125297722, 10.2328175845497, 10.4712208696213, 10.7151748001665, 10.9646022666217, 11.2201783962418, 11.4818755272534, 11.7487660694211, 12.022458611423, 12.3029582831401, 12.5892293159785, 12.8821305829416, 13.1826882436444, 13.4897550794907, 13.8044007269437, 14.1252831922455, 14.4548879516212, 14.7918546960737, 15.1360273793202, 15.488804126609, 15.8484759238666, 16.2182817980509, 16.5965079631525, 16.9830034334309, 17.3775626417757, 17.7820985331221, 18.1966079272567, 18.6210449358712, 19.0553168235015, 19.499277325153, 19.9527202232158, 20.4183315756054, 20.8930982542792, 21.3798880657389, 21.8786989662917, 22.3894775667764, 22.9082936233792, 23.4424310548708, 23.9879960081905, 24.460383892312, 25.0326798266495, 25.6158540067528, 26.2145168077636, 26.8287704886367, 27.4529619590147, 28.0982185506974, 28.7527007922617, 29.4219449727912, 30.1126649436645, 30.8108103109316, 31.5303245328492, 32.2717738292055, 33.0185315147938, 33.7956233897451, 34.5857748541972, 35.3879391515523, 36.2219479825103, 37.0673806362718, 37.9227145659252, 38.8107470771415, 39.7202499156505, 40.6510388144964, 41.6028169233534, 42.5751549892028, 43.5674822949736, 44.5790713196024, 45.6265840882283, 46.6930952545201, 47.7774087260323, 48.8985833402475, 50.0364866202866, 51.2119943690208, 52.4025439343416, 53.6309809757702, 54.8720174783855, 56.1506397828695, 57.4677958921276, 58.8244100721625, 60.1888854751973, 61.5912822397116, 63.0321122532365, 64.5118013476349, 65.9908327291353, 67.5471812303265, 69.142876649141, 70.7313612927305, 72.4030445720195, 74.0616476165949, 75.8068914427253, 77.5886895663998, 79.4057707792604, 81.2565744543678, 83.1391628600678, 85.0512132395375, 87.0635050988493, 89.1065997043318, 91.1772654809007, 93.3575399968966, 95.475486683051, 97.7030783533799, 100.049025578562, 1e3]
        Ai = @SVector [1.0, 1.16852001022416, 1.17079775558813, 1.17549243576777, 1.18035547957409, 1.18539461708148, 1.19061868119862, 1.1960365289652, 1.20165797121848, 1.20749309515359, 1.21355290678476, 1.21984946681944, 1.22639509922126, 1.23312093381688, 1.24002947681618, 1.24713134299229, 1.25443816901358, 1.2619624592546, 1.26971741940921, 1.27771686651422, 1.28597600573522, 1.29451111209051, 1.30333920927558, 1.3124791779239, 1.32195116574899, 1.331776707716, 1.34197893269775, 1.35258350017443, 1.36361750527889, 1.37510904143418, 1.38708171703191, 1.39955296281825, 1.4125356109455, 1.42603688229788, 1.44005869074998, 1.45459892974749, 1.46965020833269, 1.48520096709046, 1.50123586514995, 1.51773493846986, 1.53467480819576, 1.55202818636721, 1.5697632073289, 1.58784565595421, 1.60623781051806, 1.62489866043229, 1.6437868179371, 1.66285738790428, 1.68206487297593, 1.70136404384849, 1.72070760504375, 1.74004892012907, 1.75934194810429, 1.77854057540207, 1.79759995294883, 1.81647718025617, 1.83513000935443, 1.8535217521019, 1.87162242214568, 1.88941255272572, 1.90687780076733, 1.92401130377593, 1.94080941457954, 1.95727409653831, 1.97340995514258, 1.98922292272048, 2.00471807179774, 2.01990253139759, 2.03478072842117, 2.04935948180341, 2.06364451072219, 2.07764156534066, 2.09135709660062, 2.10479651336051, 2.11796557822619, 2.13087108092475, 2.14351749873696, 2.15591249225738, 2.16806034482347, 2.17996751811304, 2.19163921833107, 2.20308196703137, 2.21429984014348, 2.22529984531861, 2.23608642044641, 2.24666491296, 2.25703971450002, 2.26721707265114, 2.27720216870615, 2.28699795474133, 2.29661115471612, 2.30604515456765, 2.31530390963179, 2.32439429955309, 2.33331838340306, 2.34208211949183, 2.3506898685418, 2.35914131868225, 2.36744029516954, 2.37559564395953, 2.38361262925028, 2.39149589239825, 2.39924798897435, 2.40687177159353, 2.41437445939354, 2.42175720647081, 2.42902406190093, 2.43618092346483, 2.4432285288167, 2.45017201572293, 2.45701559261216, 2.46376077865167, 2.47041383682226, 2.47697697116147, 2.48344866944727, 2.48983434058688, 2.49614062074161, 2.50236906034215, 2.50852637608938, 2.51461125832967, 2.52063267268482, 2.52658971422462, 2.5324866539671, 2.53832634427074, 2.54411360296373, 2.54985171915201, 2.55554230033366, 2.56118889604372, 2.56679533022657, 2.57236361999384, 2.57790010899007, 2.58340298087726, 2.58887894228097, 2.59432847601493, 2.59975896910812, 2.60516895357878, 2.6105638526664, 2.61594465116197, 2.62131162282258, 2.62667189090808, 2.63202999377162, 2.63739013876162, 2.64275403769574, 2.6481263238693, 2.65350897203928, 2.65890396728779, 2.66431975476579, 2.66975569031906, 2.67521757147602, 2.68070832446159, 2.6862308772798, 2.69178829914687, 2.69738385815032, 2.703025010902, 2.70871579015758, 2.7144561228457, 2.72024997289303, 2.72611086425046, 2.73202821223489, 2.7380150018984, 2.74407462239375, 2.75021246773453, 2.75643435613377, 2.76274653322121, 2.76914970570235, 2.77565668059431, 2.78226217154989, 2.78897992990252, 2.79581117814499, 2.80276420246199, 2.80984800738925, 2.81706434550368, 2.82441438827793, 2.83191675510894, 2.83956602158246, 2.84729329742131, 2.85505124280977, 2.86281624903587, 2.87061255837141, 2.87841327596513, 2.88624430284398, 2.89408887463845, 2.90195206115551, 2.90983974040536, 2.91773332989093, 2.92565126054603, 2.93358774421819, 2.94155024897269, 2.94951865660694, 2.95751467858713, 2.96551741714752, 2.97355058849505, 2.98160897361581, 2.98966977885218, 2.9977591619403, 3.00585374518061, 3.0139820603, 3.02213920730227, 3.030300031461, 3.03847616292749, 3.04668114964609, 3.05488803711364, 3.06313216351745, 3.07138615416919, 3.07966556741002, 3.08796430757199, 3.09627552139215, 3.10461727824668, 3.1129839706072, 3.12134204688819, 3.12973759800806, 3.13816589795524, 3.14659154116638, 3.15503581308473, 3.16352345588758, 3.17201745225504, 3.1805430287924, 3.18905965203779, 3.1976286540199, 3.20620987311826, 3.21479525292585, 3.22341511426585, 3.23202358553016, 3.24069342682608, 3.24937927136952, 3.25807342197179, 3.26676729053254, 3.2754984735271, 3.28426181654091, 3.29305153896602, 3.3018610313924, 3.31068286760127, 3.3195087452895, 3.32838606965378, 3.33725279004637, 3.3461580434439, 3.35509655927376, 3.36406242922311, 3.37298303508608, 3.38197912197677, 3.39097974086048, 3.39861081876226, 3.40769614800642, 3.41676669713791, 3.42588721246344, 3.43505355499458, 3.44417723770067, 3.45341565746253, 3.46259423551111, 3.47178723635246, 3.4810808152818, 3.49028183382626, 3.49956986722198, 3.50894508089, 3.51819394054556, 3.52762120844497, 3.53701090587155, 3.54634776514436, 3.55585653517976, 3.56529847733069, 3.57465489115539, 3.58416958708424, 3.59371498600585, 3.60328401161063, 3.61286860245501, 3.62245992941058, 3.63204804986045, 3.64162183958739, 3.65133273396139, 3.66101764112691, 3.67066253687125, 3.68043122375501, 3.69014283075682, 3.69997010038884, 3.70971955082135, 3.71957348975725, 3.72932468594816, 3.7391652025581, 3.74909504558002, 3.75911382258568, 3.76898484207798, 3.77892230726438, 3.78892335989724, 3.7989844773555, 3.80883512598476, 3.81898849111315, 3.82918672185962, 3.83913128373885, 3.8493826833476, 3.8593459041022, 3.86961490702734, 3.87988549633358, 3.89014608065431, 3.90038312046918, 3.91058256858265, 3.92072913548489, 3.9311894664449, 3.94159368252546, 3.95192307260851, 3.96257802419261, 3.97271726066907, 3.98316253500384, 3.99393943022028, 3.99393943022028]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_bssa14_760 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Boore _et al._ (2014) GMM, for Vs30 = 760 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_bssa14_760 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_bssa14_760()
        fi = @SVector [0.001, 0.0911162626259868, 0.093260321030032, 0.0954548276537386, 0.097700983456623, 0.0999999820952542, 0.102329288657437, 0.104712886897726, 0.107151898576338, 0.109647795949619, 0.112201818697888, 0.114815374887065, 0.117489775491468, 0.120226401015101, 0.123026884861295, 0.125892513777469, 0.128824911618647, 0.131825678122484, 0.134896268392557, 0.138038384735951, 0.141253709921094, 0.144543963437463, 0.147910800512607, 0.151356075188123, 0.154881660174717, 0.158489267502305, 0.162180970646285, 0.165958653333, 0.169824346053222, 0.173780044393326, 0.177827942042926, 0.181970051029366, 0.186208660232139, 0.190546059784457, 0.194984353229929, 0.199526128106147, 0.204173751722764, 0.20892955028489, 0.213796152511466, 0.218776112752252, 0.223872048825969, 0.229086703225714, 0.234422832723251, 0.239883195800222, 0.245470813347372, 0.25118851160021, 0.257039461393428, 0.263026692540031, 0.269153371763333, 0.275422819407075, 0.281838165047044, 0.288402948293457, 0.295120803175706, 0.301995031395083, 0.309029397959102, 0.316227608503954, 0.323593446557635, 0.331130969998632, 0.338843906713985, 0.346736597226246, 0.354813252116192, 0.363077901142721, 0.371535031156678, 0.380189215767494, 0.389044984220145, 0.398106816830357, 0.407380136811554, 0.416869160401731, 0.426579333548684, 0.436515548592194, 0.446683272378262, 0.45708795418994, 0.467734860095858, 0.478629668696382, 0.489778445447913, 0.501186718043336, 0.512861077069637, 0.524807326731799, 0.537031164483197, 0.549540349983891, 0.562340880014143, 0.575439553575671, 0.58884341827149, 0.602558931622333, 0.61659402334075, 0.630956723964703, 0.645653291142765, 0.66069312756252, 0.676082650590049, 0.69183056836225, 0.707945076523136, 0.724434752191284, 0.741308974581602, 0.758576209627959, 0.776246773030056, 0.79432748801736, 0.812829309782221, 0.831763442447971, 0.851137724527408, 0.870962906834318, 0.891249503642188, 0.912009605000737, 0.933252826456988, 0.954990037596422, 0.977236096699494, 0.999997856190628, 1.02329086792856, 1.04712732847849, 1.07151799175662, 1.09647479135, 1.12201789771688, 1.14815282104044, 1.17489622351163, 1.20226311696289, 1.23026620148268, 1.25892388582697, 1.28824778446107, 1.31825530797311, 1.34896087152802, 1.38038032106668, 1.41253692508557, 1.445439022341, 1.47910778317774, 1.51355422248531, 1.54881134487854, 1.58488795476428, 1.62180838171325, 1.65958368858436, 1.69824238595341, 1.73779847972394, 1.77827603600399, 1.81969150631633, 1.86208329217625, 1.90545034897491, 1.94983902731283, 1.9952507815335, 2.04172779890817, 2.08928355129046, 2.13795963836794, 2.18775727215626, 2.23870620493322, 2.29085569085141, 2.34422606728919, 2.39881914486294, 2.45469203673004, 2.51186859307884, 2.57039504560511, 2.63025420967004, 2.69151733278799, 2.75421557941965, 2.81838194026748, 2.88402339612077, 2.9512028216528, 3.0199292355183, 3.09027425274607, 3.16225046034222, 3.23590414452624, 3.31128592143274, 3.38841127459114, 3.46733561384397, 3.54812049289905, 3.63074056844326, 3.71530742348746, 3.8018467209065, 3.89043759863809, 3.98105454269997, 4.07378124969755, 4.16864907801435, 4.26575643477429, 4.36514527255278, 4.46678500319695, 4.57086520270523, 4.67728048823422, 4.78624047019078, 4.89771957141816, 5.01186900211919, 5.12856311530136, 5.2480628160809, 5.37023313546781, 5.49536746546279, 5.62332291499947, 5.754303052447, 5.88840999817014, 6.02547308113461, 6.1658736760049, 6.30943605583888, 6.4564762291969, 6.60681845059276, 6.7607717254324, 6.918312929425, 7.07936588393754, 7.24425747015809, 7.41292207965159, 7.58574578518812, 7.76242899518572, 7.94312555149321, 8.12829407379905, 8.31759010959621, 8.5112050404209, 8.70936112698272, 8.91231611946591, 9.12000099309567, 9.33231502050119, 9.54994079660182, 9.77239944707185, 10.0000001597799, 10.2326263405644, 10.471136803066, 10.7149551224552, 10.9644882366261, 11.2202174216049, 11.481450273086, 11.7486508677961, 12.0223720914911, 12.3025291219287, 12.5889952804829, 12.8824194005891, 13.1827051012315, 13.4897101260208, 13.8032410529247, 14.1250867438041, 14.4541781486353, 14.7903148709324, 15.1356308586337, 15.4876465864803, 15.8486432543548, 16.2171642979246, 16.5958596666274, 16.9816664566089, 17.3774298143529, 17.7831140008936, 18.1967920767022, 18.6199724968074, 19.0544793670653, 19.498175111712, 19.9529754896924, 20.4163755211705, 20.892912346942, 21.3800507670956, 21.8774716691975, 22.3877472701828, 22.9076855975856, 23.4400018755962, 23.9879916115976, 24.4597343241965, 25.0308911841329, 25.6135859625918, 26.2116059288632, 26.824939523369, 27.4535021675955, 28.0920598160859, 28.7502419903201, 29.4228448805403, 30.1093798910821, 30.8092265190817, 31.5282456680064, 32.2665483915277, 33.0167888291429, 33.793334298587, 34.5807677450142, 35.3862124015741, 36.2092233431454, 37.0588715126563, 37.9257367673275, 38.8089638768307, 39.7188586552412, 40.6439586236746, 41.5954881087888, 42.5604131037149, 43.5650231077962, 44.5818544068736, 45.6085878997718, 46.675623757658, 47.7676222037955, 48.884049300592, 50.0241840099151, 51.1870940695605, 52.3932632105272, 53.6218239035964, 54.8712133818895, 56.1395620572865, 57.451503152175, 58.8086644682486, 60.1827801874379, 61.5707814890962, 63.0024208862257, 64.4788171831622, 66.0011049249864, 67.5311883608651, 69.105169857874, 70.7236199043905, 72.3869859093085, 74.0470212505932, 75.7986325153462, 77.5413587183516, 79.3804361813949, 81.2035291512659, 83.1273218954462, 85.093198969761, 87.0288650268165, 89.0705440454914, 91.1483073695668, 93.2592711043505, 95.4876919544843, 97.75081441754, 99.9469401041666, 1e3]
        Ai = @SVector [1.0, 1.08364234398491, 1.08586162583744, 1.08814730526576, 1.09050179138363, 1.09292756623953, 1.09540190454842, 1.0979513931251, 1.10057866656408, 1.10328683758114, 1.10607887745956, 1.10895805670094, 1.11192766354774, 1.11499119483819, 1.11815257551917, 1.12141549472601, 1.12478418772514, 1.12826304356699, 1.13185647450966, 1.13556938629872, 1.13940688690809, 1.14337436963043, 1.14747740913594, 1.15172210207447, 1.15611487583144, 1.16066228586858, 1.16537169899678, 1.17025063312829, 1.17530721624628, 1.18054999059327, 1.1859882627106, 1.19163164309922, 1.19749071384348, 1.20357671924515, 1.20990135521117, 1.21647779168358, 1.22331975755045, 1.23044186720427, 1.23786023490562, 1.24559202968568, 1.25365580305696, 1.26207173145057, 1.27086160427698, 1.28004897777845, 1.28965982494954, 1.29972199871975, 1.3102664347014, 1.32132661705483, 1.33293942599418, 1.34514560498503, 1.35798957118111, 1.37152116257721, 1.38579572462222, 1.40076722859397, 1.4157251995136, 1.43055370729663, 1.44522537197278, 1.45971823971733, 1.47401401012578, 1.48809929035606, 1.50196367821462, 1.51559921109471, 1.52900107663852, 1.54216660353197, 1.5550946936694, 1.56778559787145, 1.58024206161631, 1.59246616999708, 1.60446241798237, 1.61623515256263, 1.6277895136788, 1.63913048246185, 1.65026272448572, 1.66119120826138, 1.67192117847526, 1.68245722468501, 1.69280480837492, 1.70296852990295, 1.71295281949093, 1.72276364578743, 1.73240541599026, 1.74188303497685, 1.75120138626003, 1.76036482936819, 1.76937853375415, 1.77824754652269, 1.78697561303618, 1.79556821706716, 1.80402894254375, 1.81236248801982, 1.82057312873887, 1.82866517275189, 1.8366317010579, 1.84448535974719, 1.85223437956147, 1.85988175704966, 1.86743216710123, 1.87489020709522, 1.88225899266416, 1.88954271455607, 1.89674532185353, 1.90387123546318, 1.91092387874021, 1.91790706239415, 1.92482573751064, 1.93168227941724, 1.93848164305175, 1.9452276783545, 1.95192374244991, 1.95857353251229, 1.96518281160841, 1.97175341983825, 1.9782900454899, 1.98479687919082, 1.99127750153894, 1.99773677421567, 2.00417793311796, 2.01060551046604, 2.0170233256773, 2.02343547276264, 2.02984749436382, 2.03626191085848, 2.04268353875561, 2.049114602113, 2.05556084771603, 2.06202287844052, 2.0685036548133, 2.07500332318476, 2.08152466608446, 2.08806765067701, 2.09463361199228, 2.10122238518265, 2.10783707036405, 2.11447432622988, 2.12113499884932, 2.12781424713384, 2.13451956963627, 2.14124951350918, 2.14800661022883, 2.15478765632886, 2.16159339682423, 2.1684271104748, 2.17528816557151, 2.18217362577078, 2.18908761935105, 2.19603002639163, 2.20300351368397, 2.21000319758685, 2.21703470694882, 2.22409898451505, 2.2311972221775, 2.23832781958321, 2.24549538117795, 2.2526987664562, 2.25994346481583, 2.26722888065824, 2.27455793921908, 2.28193411358852, 2.28935753459416, 2.29683236874686, 2.3043633905498, 2.31194745870387, 2.31959420135704, 2.32730538253537, 2.33508774151048, 2.34293887835729, 2.35086603625359, 2.35887211445266, 2.36696575744595, 2.37515092948982, 2.38342582596344, 2.39180677070733, 2.40028643357407, 2.40888278386229, 2.41759536090252, 2.42643775463242, 2.43540229532395, 2.44451132297608, 2.45375713864841, 2.46316480256067, 2.47272675848702, 2.4824616712993, 2.49238096112596, 2.50247621182259, 2.51278005644824, 2.52328459530469, 2.53400459669908, 2.54482080340439, 2.55569228691112, 2.56660276366681, 2.57754142971351, 2.58852482518397, 2.59954284066688, 2.61061488013636, 2.62171598016116, 2.63285030329271, 2.64404020662793, 2.65525905027419, 2.66651256824853, 2.6778079007294, 2.68915376033145, 2.70054028705822, 2.7119562393715, 2.72343228376484, 2.73493710925336, 2.74648102444296, 2.75805235043705, 2.76968764453411, 2.78135267969627, 2.79306102505984, 2.80482891692235, 2.8166185142729, 2.82844500757262, 2.84032661392745, 2.85225330543254, 2.86421363347959, 2.8762285042368, 2.88828754085377, 2.90037895881915, 2.91248939932882, 2.92468148368883, 2.93690801495623, 2.9491556390582, 2.96149548540931, 2.97383256600454, 2.98624067565195, 2.99866340382951, 3.01118335725144, 3.02369287827608, 3.03627804951518, 3.04893057785931, 3.06158415136629, 3.07427892426899, 3.08706261807415, 3.09986553992361, 3.11273641536514, 3.12559813082232, 3.13856985463709, 3.15157524789462, 3.16459965605965, 3.17770366601632, 3.1907987943681, 3.20394736944601, 3.21722278176947, 3.22842357616881, 3.24176833114294, 3.25512378677881, 3.26856595639597, 3.2820866712014, 3.29567642336465, 3.30921614709875, 3.32290298108548, 3.33662052652846, 3.35035255916578, 3.36408076000853, 3.37791285360432, 3.39184244115442, 3.40572446603041, 3.41981670509913, 3.43383179087081, 3.44789098701598, 3.46197960670245, 3.47624422758958, 3.49051829584613, 3.5047817650066, 3.51919309288275, 3.53356349271198, 3.54806006042791, 3.5624775690411, 3.57719953201365, 3.59181448730171, 3.60628635222813, 3.62103589912565, 3.63584052451261, 3.65068559075673, 3.66555440665079, 3.68042816483525, 3.69555904026841, 3.71067592274549, 3.72575410093648, 3.74076605300308, 3.75599484014451, 3.77144717971867, 3.7867934023843, 3.80199625298696, 3.81737457761158, 3.83293031529568, 3.84866317013457, 3.86417349493963, 3.87982169469459, 3.89560324156142, 3.91151251638117, 3.92708445960429, 3.94319976493462, 3.95892568992393, 3.97520292066274, 3.99102935641841, 4.00740962554074, 4.02383037958937, 4.03968828470414, 4.05609239283177, 4.07246729952653, 4.0887846867556, 4.10568188902212, 4.1225184203158, 4.13853906719252, 4.13853906719252]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_bssa14_1100 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Boore _et al._ (2014) GMM, for Vs30 = 1100 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_bssa14_1100 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_bssa14_1100()
        fi = @SVector [0.001, 0.0999999942707916, 0.102329289439609, 0.104712887716761, 0.107151912555722, 0.109647810587833, 0.11220183402598, 0.114815390937559, 0.117489792298401, 0.120226418614101, 0.123026885991878, 0.125892533074394, 0.128824931825004, 0.131825679420567, 0.134896290548396, 0.138038407935955, 0.141253711030388, 0.14454397891139, 0.147910809014606, 0.151356098582797, 0.154881696694555, 0.158489297243423, 0.162180974138637, 0.165958674331861, 0.169824392157856, 0.173780066819787, 0.177827977607297, 0.181970093315908, 0.186208684639896, 0.190546064687612, 0.194984400298219, 0.199526151255418, 0.204173764851596, 0.208929596943051, 0.213796152899848, 0.218776165415316, 0.223872062165998, 0.229086795091157, 0.234422839210589, 0.239883288165173, 0.245470839468972, 0.251188557388039, 0.25703952116282, 0.263026714586751, 0.269153406683434, 0.275422798880656, 0.281838241124289, 0.288403025855549, 0.295120803536784, 0.301995173753911, 0.309029502480207, 0.316227782943992, 0.323593500466634, 0.331130953508207, 0.338844036623896, 0.34673675462895, 0.354813392739313, 0.363078026106895, 0.371535181708974, 0.380189334311823, 0.389045073797216, 0.398107089497488, 0.407380173586767, 0.416869210701438, 0.426579417203211, 0.43651559882169, 0.446683401771556, 0.457088029401324, 0.467735011805406, 0.478629925081276, 0.489778715181216, 0.501187031937496, 0.512861236791132, 0.524807370945154, 0.537031434094109, 0.549540627902458, 0.562340918238859, 0.575439503820714, 0.588843301006728, 0.602559271619929, 0.616594838924024, 0.630956918926108, 0.645653987282243, 0.660693278719166, 0.676082386262125, 0.691830704330551, 0.707945463071725, 0.724435696155354, 0.74131024499222, 0.758576833362888, 0.776246967350283, 0.79432765983909, 0.812829847543139, 0.831763295436004, 0.851137257381919, 0.870963647695458, 0.89125063310826, 0.912010190336213, 0.933253876719992, 0.954991323260685, 0.977236674564363, 0.999999376704014, 1.02329216650515, 1.04712858016108, 1.07151802029997, 1.09647705386315, 1.12201813106039, 1.14815232082574, 1.17489711452937, 1.2022632761053, 1.23026893465442, 1.25892478169571, 1.28824925286246, 1.31825476678541, 1.34896226936176, 1.38038331625282, 1.41253536157361, 1.44544002786044, 1.47910755205328, 1.51355783588285, 1.54881556831327, 1.58489238265031, 1.62180711433383, 1.65958346288574, 1.69824231541703, 1.73780089595023, 1.77827646757433, 1.81969641187082, 1.86208481433417, 1.90545626586434, 1.94984475199332, 1.99526162858629, 2.04173373924813, 2.08929105688858, 2.13796119637597, 2.18776046050659, 2.2387215306585, 2.29086418676164, 2.34422580916097, 2.3988291206262, 2.45470634410607, 2.51188114589687, 2.57038711926446, 2.63025902311945, 2.69153260246708, 2.75421969068968, 2.81837995348003, 2.88402365216443, 2.95119935757395, 3.01994331510937, 3.09029180950475, 3.16226398102817, 3.23592925568545, 3.31130709484967, 3.38843294207687, 3.46736315436668, 3.54813581847585, 3.6307651436782, 3.71533568298211, 3.80188946845371, 3.89044115666823, 3.98105964067485, 4.07378964803908, 4.16867752173957, 4.26577171067028, 4.36516053182719, 4.4668254158747, 4.57086030476262, 4.67732851465, 4.78629925533731, 4.89774978439209, 5.01185370947343, 5.12859423484337, 5.24805927489683, 5.37028645796196, 5.49537951621125, 5.6233878102801, 5.75436328712337, 5.88843917410242, 6.02560553448902, 6.16592821030288, 6.30957217415461, 6.45653608003565, 6.60691210897118, 6.760804602669, 6.91833243844884, 7.07938781287924, 7.24434482940889, 7.41310826719381, 7.58569409562697, 7.76241731627088, 7.94317188337215, 8.12832595039509, 8.3176023537528, 8.51140752074882, 8.70962908279862, 8.91253998805093, 9.12001705818987, 9.33259809073417, 9.54976840803428, 9.77223833172881, 10.0000154584674, 10.23300524141, 10.4710754835424, 10.7150266728374, 10.9647907700409, 11.2202630001957, 11.4812971912249, 11.7489102165068, 12.0226227359966, 12.3027246006294, 12.5890770378463, 12.8824938912425, 13.1823630974293, 13.4895877351951, 13.8034850183452, 14.1250621277036, 14.4542171297727, 14.7907986463695, 15.1353290208207, 15.4876534354551, 15.8483711149959, 16.2181900174972, 16.5961299741433, 16.9819204095141, 17.37824271703, 17.7830169525712, 18.1970367629386, 18.6200661560741, 19.0542962143714, 19.4984106230298, 19.952167919195, 20.4166708483612, 20.8932695398002, 21.3786496995455, 21.8773864462395, 22.3877335357594, 22.90750432443, 23.4421462082247, 23.9875431882539, 24.4595049310709, 25.0319269878903, 25.6166097703306, 26.2130779475033, 26.8263113471932, 27.4535149251717, 28.0943122232752, 28.7514900856313, 29.4214814188544, 30.1071964789614, 30.8121657464104, 31.5324826105913, 32.2675988220233, 33.0213566963618, 33.7936159040409, 34.579074170418, 35.3872354433233, 36.2127668867604, 37.0609573167903, 37.9256441289802, 38.8124796011373, 39.7213308469351, 40.6445318091086, 41.5961301188189, 42.5685278399345, 43.5610008566936, 44.5726281197036, 45.6216075347537, 46.6789911567352, 47.773980830181, 48.8853621422687, 50.0351078965322, 51.1992803651972, 52.3886636425958, 53.6164785697481, 54.8689227094684, 56.1447051106333, 57.4588446996262, 58.8120630745198, 60.1679954909794, 61.5799079121313, 63.0304154891204, 64.497885722967, 66.0018126389695, 67.5418009953363, 69.117236380239, 70.7272485242521, 72.3706846807164, 74.0761580253903, 75.8148416156396, 77.5847393108404, 79.3834328727271, 81.245390419232, 83.1336688348237, 85.0861844185969, 87.0612501896228, 89.1005448083237, 91.1570157956521, 93.3280977544332, 95.4597653710987, 97.7091485760571, 99.9714285714285, 1e3]
        Ai = @SVector [1.0, 1.07377424172191, 1.07568340471177, 1.07764765689837, 1.07966876479455, 1.08174881857823, 1.08388980782557, 1.08609391640319, 1.08836331477285, 1.09070030404772, 1.09310746523746, 1.09558722513272, 1.0981422996958, 1.10077552311629, 1.10348975891098, 1.10628809163659, 1.10913843677795, 1.1119862441656, 1.11482032357681, 1.11763348969258, 1.12041986774419, 1.12317468498611, 1.12589456554055, 1.12857701666666, 1.13122044460195, 1.13382399537393, 1.13638771748597, 1.13891205161871, 1.14139812609109, 1.14384729457583, 1.14626080510911, 1.14864002376794, 1.15098626482962, 1.15330074743668, 1.15558472198198, 1.15783950909849, 1.16006624659493, 1.16226627242009, 1.16444066637276, 1.16659074938093, 1.16871763839306, 1.17082256875985, 1.17290674545382, 1.17497098881197, 1.17701661304444, 1.17904472323173, 1.1810564402846, 1.18305283340685, 1.18503506676802, 1.18700424346501, 1.1889613920927, 1.19090768763071, 1.19284414873617, 1.19477198498936, 1.19669227954613, 1.19860612604854, 1.20051467448373, 1.2024189984736, 1.20432027155042, 1.20621964116529, 1.20811825260841, 1.21001727350919, 1.21191787262328, 1.21382122429834, 1.2157285626702, 1.21764102751111, 1.21955990799381, 1.22148640147232, 1.22342175358253, 1.22536721997457, 1.22732410176577, 1.22929364971439, 1.23127722537588, 1.23327612965734, 1.23529166096294, 1.23732531195325, 1.23937836912328, 1.24145231559765, 1.24354749289177, 1.24566554333833, 1.24780897539437, 1.2499792569878, 1.25217809355537, 1.25440701343445, 1.2566676191693, 1.25896178257908, 1.26129108887827, 1.26365739381181, 1.26606255846768, 1.26850829677884, 1.27099689683087, 1.27353004434842, 1.27611001171338, 1.27873893586496, 1.28141893992751, 1.28415254235385, 1.28694181625922, 1.28978939335592, 1.2926979191094, 1.29566982523915, 1.29870824175611, 1.30181572532181, 1.30499534613735, 1.30825038999295, 1.31158365403668, 1.31499901534591, 1.31849988955826, 1.32208961620671, 1.32577251378186, 1.32955179855246, 1.33343132606967, 1.33741357887289, 1.34150179466669, 1.34569812166275, 1.35000567098599, 1.3544260520849, 1.35896154743919, 1.36361490236401, 1.36838718131483, 1.37328072536624, 1.37829854584582, 1.38344177665736, 1.38871258893308, 1.39411388408896, 1.39964824530506, 1.40531779136425, 1.41112476128775, 1.41707296818823, 1.423165914765, 1.42940562445159, 1.43579526073916, 1.44233948342952, 1.44904300900007, 1.45591026098732, 1.46294377577838, 1.4701430068133, 1.47750842207663, 1.48503741802188, 1.49272885060967, 1.50057846594505, 1.50858248846468, 1.51673514116072, 1.52503138142624, 1.53346570997836, 1.54203225573951, 1.55072136469112, 1.55952965490261, 1.56844664272352, 1.57746682872857, 1.5865828595448, 1.59578726004598, 1.60507023827321, 1.61442848210392, 1.62385227018904, 1.63333398385059, 1.64286864207467, 1.65244910763165, 1.66206559908029, 1.67171669047243, 1.68139615123491, 1.69109489130831, 1.70081005746505, 1.7105363237882, 1.72026875507274, 1.73000284207063, 1.73973826625622, 1.74946440275484, 1.75918182984198, 1.76888828501048, 1.77858223152089, 1.7882542429275, 1.79791222543811, 1.80754789891303, 1.81716203507685, 1.82675150733127, 1.83631855455804, 1.84586136433489, 1.85537854764995, 1.8648747408499, 1.87434449935178, 1.88378803080827, 1.89321224044973, 1.90261311295528, 1.91199291860342, 1.92135475667402, 1.93070261245488, 1.94002726561405, 1.94934764561085, 1.95865579899324, 1.96795051189792, 1.97724668214133, 1.98653689629335, 1.99583838905708, 2.00513582401076, 2.01444803976094, 2.02376863922482, 2.03310968744784, 2.04246501089894, 2.05181609311956, 2.06121173032725, 2.07066458274414, 2.08015734029654, 2.08968114874409, 2.09922590184368, 2.10881876661101, 2.11845189129642, 2.12811625143497, 2.1378015484207, 2.14754048090738, 2.15731038499326, 2.16711663195644, 2.17694945369856, 2.18683162558877, 2.19673745095695, 2.2066916515308, 2.21666701444773, 2.22669041520284, 2.23675336485491, 2.24684609768205, 2.25697912727245, 2.26714268360788, 2.27734887295641, 2.28761212076916, 2.29789987370863, 2.30819997235961, 2.31857847380772, 2.32897531728647, 2.3394058792482, 2.3498590462889, 2.36038337973389, 2.37094068336684, 2.38151809194227, 2.3921384760639, 2.40282679650948, 2.4135035480742, 2.42426389889233, 2.43506415004308, 2.44585309227987, 2.45673813861894, 2.46762963485226, 2.47687509577523, 2.48790470398334, 2.49895459013647, 2.51001074938325, 2.52115955775645, 2.53234387408955, 2.54355153871219, 2.55482544462744, 2.56609879830649, 2.57741526963207, 2.58882656935048, 2.60026315087013, 2.61171104942407, 2.62322430649175, 2.63479438821278, 2.64633662881618, 2.65798476086959, 2.66965548300257, 2.68141714192981, 2.6931782663281, 2.70500988439125, 2.71690361145269, 2.72875549295849, 2.74073817576405, 2.7527489711529, 2.76477328510593, 2.77679473796317, 2.78902190862775, 2.80111149742784, 2.81339155369115, 2.82561739127939, 2.83802390290198, 2.85034656746604, 2.86269510132692, 2.87519926745502, 2.88771129849868, 2.90021311182686, 2.91284514811118, 2.92560594571757, 2.93814853329408, 2.95095963812371, 2.96387158874595, 2.97668665984961, 2.98957013452994, 3.00251182166991, 3.01549984515371, 3.02852071664694, 3.04155909263268, 3.05483342147564, 3.06811110345284, 3.08137173364153, 3.09459254417313, 3.10801932770893, 3.12137858766394, 3.13493150812508, 3.14838227829193, 3.16200828231978, 3.17548946061531, 3.18945462042179, 3.20290819766718, 3.21683654256236, 3.23041469686689, 3.23041469686689]
        new(construct_interpolant(fi, Ai))
    end
end


"""
    struct SiteAmpAlAtikAbrahamson2021_cb14_620 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Campbell & Bozorgnia (2014) GMM, for Vs30 = 620 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_cb14_620 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_cb14_620()
        fi = @SVector [0.001, 0.0999999960075922, 0.102329291970146, 0.104712897083861, 0.107151915826179, 0.109647820193404, 0.112201826611791, 0.11481539188404, 0.117489781977474, 0.120226413052329, 0.123026878155944, 0.125892534108239, 0.12882492604089, 0.131825673462998, 0.134896299505685, 0.138038403078452, 0.141253717665068, 0.144543977586134, 0.147910828038878, 0.1513561097844, 0.154881688910334, 0.158489270249457, 0.162180956299533, 0.165958652768952, 0.169824403593973, 0.17378007921449, 0.177827963368866, 0.181970051450291, 0.186208690172977, 0.190546100984621, 0.194984442453321, 0.199526172070992, 0.204173806745097, 0.208929601318305, 0.213796149585515, 0.218776122992034, 0.223872028310857, 0.229086734657301, 0.234422921222406, 0.23988324885248, 0.245470785537473, 0.251188484278206, 0.25703946055711, 0.263026674560631, 0.269153407076944, 0.275422773948591, 0.281838261914018, 0.288402967980864, 0.295120835838376, 0.301995143474081, 0.309029499502454, 0.31622766697645, 0.323593638658402, 0.331131118022553, 0.338844074067532, 0.346736845850547, 0.354813149043298, 0.363078172132964, 0.3715349775957, 0.380189376407706, 0.389044877600162, 0.398107205333068, 0.407380389280958, 0.416869425800305, 0.426579396739984, 0.436515457062149, 0.446683422598966, 0.457087947174548, 0.467734897047422, 0.478630212709465, 0.489778263708941, 0.501187178344085, 0.512861085254938, 0.524807227283619, 0.537031938860528, 0.549540293640881, 0.56234113214465, 0.575439992722043, 0.588842984739594, 0.602559388702931, 0.61659446082985, 0.630956660969879, 0.64565404595163, 0.660692477482551, 0.676083233360251, 0.691830448958141, 0.707945639355109, 0.724436060249336, 0.741309595360551, 0.758577006429437, 0.776245814931473, 0.794326521676393, 0.812830781572186, 0.831763706566927, 0.851138659821746, 0.870961957609269, 0.891249239427603, 0.91200862589844, 0.933252340414934, 0.954990587903553, 0.977234624287875, 1.00000079350021, 1.02329093361383, 1.04712770534758, 1.07151859370928, 1.09647590648705, 1.12201792525089, 1.14815429506382, 1.17489476400895, 1.2022611083602, 1.23026524717676, 1.25892597737345, 1.28825017880555, 1.31825834042905, 1.34895940575925, 1.38038211863746, 1.41253579162943, 1.44543802508952, 1.47910766627126, 1.51355508362719, 1.54881083941909, 1.58488860908463, 1.62180200826005, 1.65958603883503, 1.69824394199126, 1.73780316909269, 1.77828002289597, 1.81969123986389, 1.86208484018179, 1.90545082777332, 1.94984212098705, 1.99526550027925, 2.04172582192457, 2.08928518503298, 2.13795270811858, 2.18775757035363, 2.23870838137221, 2.29086039430598, 2.34422567655106, 2.39881473984119, 2.45469305825437, 2.51187628530218, 2.57037914400445, 2.63024799909587, 2.69154184700609, 2.75420699435402, 2.81836339944471, 2.88403732099006, 2.95121327088087, 3.01995632802365, 3.09029370438711, 3.16225327413674, 3.23591509294214, 3.31131632426641, 3.38843926600486, 3.46737644862356, 3.54810871326601, 3.63073920182982, 3.71532481633097, 3.80191715301573, 3.8904168412904, 3.98102517597589, 4.07381013041093, 4.16866930569122, 4.2657516979992, 4.36513308947507, 4.46679458151173, 4.57081398335729, 4.67727807509726, 4.78628382607805, 4.8978150175713, 5.01184374683731, 5.12860406183254, 5.24811501763448, 5.370354782551, 5.49541292766055, 5.62344224121309, 5.75440929301849, 5.88845028920835, 6.02552436197588, 6.16598535533968, 6.30959896434349, 6.45652784724631, 6.60695947044187, 6.76085595082843, 6.91816011314, 7.07935425017163, 7.24441981839727, 7.41300823943585, 7.58566898806598, 7.76236722449045, 7.94340960477486, 8.1283883699323, 8.31760068839723, 8.51139597978318, 8.70973668407653, 8.91256157472234, 9.12027591191496, 9.33231580217011, 9.54962715813997, 9.77218973081615, 9.99995871260699, 10.2328608720273, 10.4714644143944, 10.7150241702049, 10.9648567889945, 11.2201869719567, 11.4817058148805, 11.7484937722766, 12.0221659434902, 12.3027358475711, 12.5891758955214, 12.8823542344927, 13.1821839463439, 13.489720878884, 13.8037291969212, 14.1253071115601, 14.4543872700214, 14.79086154007, 15.1361059392185, 15.4885379957005, 15.847907430289, 16.2174688117694, 16.5955044741655, 16.9818658755626, 17.3784310156056, 17.7830619796974, 18.1977471964094, 18.6200176357248, 19.0545238754325, 19.4988048035604, 19.9526560907815, 20.4158065492293, 20.8910452127662, 21.3784189555268, 21.8779299082402, 22.3858674720414, 22.9092727160828, 23.4404201902772, 23.9869860567893, 24.4596396776056, 25.033196582825, 25.6177357509289, 26.2127288715371, 26.8284448979696, 27.454199013626, 28.0951754246842, 28.7512564075941, 29.4222442059403, 30.1078484229437, 30.807682121848, 31.5290964772775, 32.2643964180019, 33.0215786275117, 33.7919505340467, 34.5842004039283, 35.3885014295223, 36.2142816944529, 37.0616880861435, 37.9308016109299, 38.8091308114225, 39.7209573241117, 40.6403219683072, 41.5939194340003, 42.5680815430318, 43.5622048396415, 44.5755162204328, 45.6249575267801, 46.6744166561296, 47.7792678540101, 48.881282995356, 50.0196194346816, 51.1955716818627, 52.3860799284434, 53.6143917622613, 54.8816520551667, 56.1605432543999, 57.4478309651699, 58.8028448839455, 60.1646294174391, 61.5638001377288, 63.03757035448, 64.4757104760039, 65.9891828579598, 67.541089429508, 69.1313510023053, 70.7121318222854, 72.3758936550096, 74.076537038047, 75.8131413932449, 77.5844999532611, 79.3891069111205, 81.225118605628, 83.1585169339083, 85.0534289506107, 87.0474892610037, 89.0690364359942, 91.1982199782074, 93.2671864620731, 95.4439739205107, 97.7371972733264, 99.9502534854245, 1e3]
        Ai = @SVector [1.0, 1.57033405486421, 1.57928537133846, 1.5882443552979, 1.59719965835096, 1.60615037691665, 1.61509589890235, 1.62403759131615, 1.63297715737857, 1.64191774399028, 1.65086388085234, 1.65982006375264, 1.66879202990905, 1.67778637671133, 1.68680976820965, 1.69586826072783, 1.70496711026745, 1.71411049690332, 1.7233013788726, 1.73254246374991, 1.74183579026946, 1.75118236586368, 1.7605837096383, 1.77004039615484, 1.77955290426151, 1.78912098882984, 1.79874517521443, 1.80842472278825, 1.81815838538661, 1.8279431990001, 1.83777482204266, 1.847648503946, 1.85755867843565, 1.86749840056441, 1.87746076155485, 1.88743842572507, 1.89742324314167, 1.90740740560039, 1.91738236364302, 1.92733937216075, 1.93727021584822, 1.94716638283987, 1.95701954525294, 1.96682109681485, 1.97656301068436, 1.9862370314421, 1.99583558310959, 2.00535062368774, 2.0147754565186, 2.02410291471308, 2.0333271817398, 2.04244361570107, 2.05144870568788, 2.06033931876222, 2.06911334678662, 2.0777696281856, 2.08630686305669, 2.09472576134762, 2.10302521344491, 2.11120721842349, 2.11927181347662, 2.12722134686463, 2.13505684751389, 2.14278036248588, 2.15039416362452, 2.15790068229876, 2.16530297151385, 2.17260321315676, 2.17980459911455, 2.18691032613591, 2.19392264524007, 2.20084615356824, 2.20768298715709, 2.21443711451096, 2.22111186161534, 2.22770982191876, 2.23423547911001, 2.24069159042248, 2.24708108770116, 2.25340840278104, 2.25967593176321, 2.26588752258148, 2.27204672331724, 2.27815610589955, 2.28422037781563, 2.2902413531033, 2.29622360202804, 2.30217007026641, 2.3080838666615, 2.31396904022813, 2.31982851395299, 2.32566610644972, 2.33148596514866, 2.33729013362996, 2.34308321614512, 2.34886759750601, 2.35464841876891, 2.3604284216875, 2.36621096026657, 2.37199820664638, 2.37779200513412, 2.38359500514961, 2.38940563224986, 2.39522716042303, 2.40105877713631, 2.40690047817613, 2.41275338473018, 2.4186163420362, 2.42448810932507, 2.43036984728744, 2.43626037019897, 2.44215988899846, 2.44806594190473, 2.45397887637024, 2.45989462958907, 2.46581733300886, 2.47174498373146, 2.4776771502338, 2.48361352566415, 2.48955235871665, 2.49549539168965, 2.50144029099867, 2.50738476075224, 2.51333259842791, 2.51928106312458, 2.52523129344264, 2.53118274865988, 2.5371349827321, 2.54309212911242, 2.54905001572751, 2.5550133144793, 2.56098043050294, 2.56694957836074, 2.57292648938209, 2.57890996786651, 2.58490147540621, 2.59089995904754, 2.59691000574735, 2.60293102740605, 2.60896245116745, 2.61500977726282, 2.62107303848077, 2.6271522095401, 2.63325068832288, 2.63936831634584, 2.64550058510771, 2.65166151868542, 2.65785253051351, 2.66407117592585, 2.67032266126765, 2.67660876019877, 2.6829313842409, 2.68929710317829, 2.69570878361406, 2.70216475706617, 2.70867278351802, 2.71523137759699, 2.72184913944897, 2.72852724055432, 2.73527126533899, 2.74207639150456, 2.74895873025397, 2.75592417006905, 2.76296587824797, 2.77009595428481, 2.77732099249509, 2.78464089712506, 2.79206268905525, 2.79959408419579, 2.80724358071353, 2.81501183363214, 2.82289893456872, 2.83092330320293, 2.83907582550249, 2.84730018640715, 2.85555328637664, 2.86383095493905, 2.87212663469594, 2.88044452184487, 2.88877774715998, 2.89714331231447, 2.9055228463282, 2.91392149979111, 2.92234551866876, 2.93078838320923, 2.93924265146354, 2.94772982803101, 2.95624421000921, 2.9647634804854, 2.97331109764328, 2.98188075759158, 2.99048268860332, 2.99909303455435, 3.00772136112323, 3.01637899018538, 3.02505962628115, 3.0337560497031, 3.04248113777928, 3.0512068177085, 3.05996757109523, 3.06875768335127, 3.07757061462058, 3.08639905796099, 3.09525984142472, 3.10412083889499, 3.11302529099264, 3.12194082614448, 3.13088687907477, 3.13982764732355, 3.14881265843947, 3.15783705140395, 3.16686309421443, 3.17591375916796, 3.18498164952413, 3.19409373759506, 3.20320853771863, 3.21235342726676, 3.22152158334157, 3.23070531974662, 3.23993702000599, 3.24916962731165, 3.25839259184477, 3.26768430453487, 3.27699608714072, 3.28631970927527, 3.29569529879741, 3.30506739375089, 3.31447732989456, 3.32386468982047, 3.33332789297788, 3.34280760438601, 3.3522949155104, 3.36177981099647, 3.37131453562767, 3.38089422257278, 3.39051328314514, 3.40009606457532, 3.40977020114658, 3.41938819505795, 3.42908426997333, 3.43729703928329, 3.44708996317642, 3.45687126433619, 3.46662471908397, 3.47651254297438, 3.48635738496203, 3.49623660010644, 3.50614293449502, 3.5160682080079, 3.52600342040139, 3.53593824695989, 3.54597106668068, 3.55598908309274, 3.56609567812751, 3.57616939738255, 3.58631878730317, 3.59641309240748, 3.60656607584072, 3.61677334812992, 3.62702952128797, 3.6371837435942, 3.64751090270576, 3.6577117152396, 3.66807717840373, 3.67845146777531, 3.68882335291022, 3.6991805679675, 3.70968940062124, 3.71998474093807, 3.73060326020412, 3.74097956442553, 3.75147967527224, 3.76210655840206, 3.77264675104285, 3.78330083425741, 3.79407041521322, 3.80471878819626, 3.8152187962405, 3.82604651276174, 3.83670771854379, 3.8474388315386, 3.85851371609353, 3.86910186430903, 3.88001792479987, 3.89098475034529, 3.90199566072941, 3.91271860783396, 3.92377506043213, 3.93484827303173, 3.94592700775264, 3.95699898517208, 3.96805033005292, 3.97906573257979, 3.99043107177771, 4.00134459652396, 4.0125958139824, 4.02377110140484, 4.03530425374715, 4.04628442747604, 4.05760168138967, 4.0692845999532, 4.08033135924049, 4.08033135924049]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_cb14_760 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Campbell & Bozorgnia (2014) GMM, for Vs30 = 760 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_cb14_760 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_cb14_760()
        fi = @SVector [0.001, 0.100000014603706, 0.102329318159967, 0.104712915012989, 0.107151944005561, 0.109647834841026, 0.112201860006149, 0.114815417052971, 0.117489816119846, 0.120226465083992, 0.123026917598311, 0.12589255847292, 0.128824980611895, 0.131825718608755, 0.13489631959351, 0.138038448458978, 0.141253778162648, 0.144544046818743, 0.14791085550833, 0.15135613998006, 0.154881728490947, 0.158489361043098, 0.162181066114414, 0.165958739473193, 0.169824462314567, 0.173780134007044, 0.177828057012465, 0.181970178254594, 0.186208750403258, 0.190546170351366, 0.194984503373748, 0.199526253525846, 0.204173881532292, 0.208929659763176, 0.213796267563971, 0.218776303013584, 0.223872214384926, 0.229086862026185, 0.23442302280373, 0.239883411745978, 0.245470999998042, 0.251188676985345, 0.257039624252806, 0.263026937897839, 0.269153625701278, 0.27542302191172, 0.281838464424222, 0.288403301147695, 0.29512107452525, 0.301995393129344, 0.30902977845631, 0.316228002125454, 0.323593765761453, 0.331131283999689, 0.338844310340282, 0.346736913437449, 0.354813558809766, 0.363078396657601, 0.371535418861104, 0.380189571348915, 0.389045434107864, 0.398107399490668, 0.407380657208765, 0.416869685071214, 0.42657993322434, 0.436516004536019, 0.446683859270794, 0.457088421802735, 0.46773557493424, 0.478630341437807, 0.489779219952398, 0.501187729749159, 0.51286184118858, 0.524808020308976, 0.537032011051797, 0.549541310846063, 0.562342002103819, 0.575440541048431, 0.588843839501076, 0.60256016086967, 0.616595892595373, 0.630957770918203, 0.64565488528681, 0.660694099005814, 0.676083721255284, 0.691831588655876, 0.707947245243919, 0.724437261196506, 0.741311001398856, 0.758578542847393, 0.77624787433371, 0.7943289003767, 0.812832326143451, 0.83176463922089, 0.851139588339233, 0.87096487176409, 0.89125241684504, 0.912013175488765, 0.933256759031112, 0.954993351540294, 0.977238752862432, 1.00000281507724, 1.02329589536783, 1.04713183702237, 1.07151977051983, 1.09647863942794, 1.12201969918408, 1.14815484858075, 1.17490036562259, 1.20226644031, 1.23027166414146, 1.25892824290883, 1.28825410786601, 1.31826027308295, 1.3489636293766, 1.38038825747881, 1.41253934332401, 1.44544429134379, 1.47910969949534, 1.51356676547177, 1.5488237651326, 1.58489501625938, 1.62181751695932, 1.65959380063872, 1.6982485957995, 1.73780868148322, 1.77828487949496, 1.81970612142534, 1.86209381862123, 1.90547035876538, 1.94984787932165, 1.99527276508788, 2.04174773586109, 2.08929818858466, 2.1379648707227, 2.18776405288898, 2.23872658201823, 2.29086951510372, 2.34424314082561, 2.39885056394356, 2.45470985693216, 2.51189788648169, 2.57039937860281, 2.63027865661929, 2.69153980303586, 2.75423167942888, 2.81840878980984, 2.88405112983486, 2.95121458203317, 3.01996235953973, 3.09030294817901, 3.16230771221273, 3.23595279682673, 3.31131420269854, 3.38847875979375, 3.46738234844352, 3.54815630214939, 3.63081856141197, 3.71538457921618, 3.80191738456751, 3.89048765512902, 3.9811191290653, 4.07383423833312, 4.16871555438765, 4.26585690197907, 4.36516008275251, 4.46685890411257, 4.57092745527101, 4.67740612206366, 4.78633645892996, 4.89784971892356, 5.01191184077053, 5.12866862527161, 5.24808342973394, 5.37031988154554, 5.49545434590912, 5.62345032003805, 5.75450629367736, 5.88845682702725, 6.02564882727567, 6.16605471650829, 6.30963041580404, 6.45668342983238, 6.60703911972332, 6.76082665306519, 6.91837183250296, 7.07962072387972, 7.24449391639376, 7.41310687103337, 7.58582639814176, 7.76259086720391, 7.9433108188294, 8.12841975058345, 8.3178495102927, 8.51150146777084, 8.70989612339629, 8.91262221663962, 9.12025383033394, 9.33271197764874, 9.54988323105152, 9.7724721984205, 9.99996812811373, 10.2331611680284, 10.4714887348519, 10.7153094829665, 10.9650477156109, 11.2206102509667, 11.4818638675441, 11.7492897173357, 12.0227614826089, 12.3028436308968, 12.5894276885153, 12.8823599384561, 13.183168411495, 13.4900328731246, 13.8045707148106, 14.1256810079244, 14.4552285070842, 14.7909197203426, 15.1359344891696, 15.4890678426225, 15.8488063029024, 16.2189403486686, 16.5966169802554, 16.9830373797447, 17.3779909265311, 17.7829518929512, 18.1978387440727, 18.6205542822369, 19.054714481358, 19.4981687206825, 19.9528554447325, 20.4186812309659, 20.892918427583, 21.3803477657443, 21.8781975901788, 22.3890774682986, 22.9097567367247, 23.4429911673855, 23.9885714919413, 24.4601011242133, 25.0325539431898, 25.6166830021738, 26.2163678794803, 26.8269705537117, 27.4525567236701, 28.0980792478464, 28.7531530453907, 29.4222858562447, 30.1110072308346, 30.8131796776573, 31.534781791216, 32.26876583596, 33.0215757243594, 33.793093567672, 34.5830953354176, 35.3912379831092, 36.2170412322082, 37.0598739432464, 37.9294021644798, 38.8152454026809, 39.7162806545213, 40.6434811545621, 41.5971967887437, 42.5639610481806, 43.5707930432651, 44.5893263185931, 45.617083905354, 46.6853663549747, 47.7782634755642, 48.8951335569009, 50.0351265950035, 51.1971650151556, 52.4023375999969, 53.6289101690096, 54.8751195274679, 56.1652603104612, 57.4731641472552, 58.8255726590189, 60.1929261276225, 61.6046338578044, 63.0272953425698, 64.4932276749624, 66.0033220481268, 67.558393760812, 69.1160679775367, 70.761078527248, 72.4046872926073, 74.0902793585822, 75.8175739578802, 77.5860562396035, 79.3949372542713, 81.2431323797365, 83.1292864603112, 85.1218876134366, 87.0813329925696, 89.1504849158889, 91.1729673616224, 93.3072493997223, 95.4704730122531, 97.7555535899792, 100.06993006993, 1e3]
        Ai = @SVector [1.0, 1.32056069012663, 1.32603400047941, 1.33153115429065, 1.33704192096276, 1.34256105272891, 1.34808429250704, 1.35360891684945, 1.3591331271623, 1.36465634908849, 1.37017920927597, 1.37570278846732, 1.38122931992161, 1.38676157440047, 1.39230292960457, 1.3978573940284, 1.40342899645319, 1.40902181854226, 1.41463965646993, 1.42028685431035, 1.42596752147355, 1.43168565135368, 1.43744569958573, 1.44325194656313, 1.44910894930871, 1.45502037208018, 1.46098871709202, 1.46701423830498, 1.47309610841515, 1.4792322941317, 1.48541907997794, 1.49165227506658, 1.49792662436597, 1.50423577746584, 1.51057319564515, 1.51693156350907, 1.5233028433285, 1.52967904845444, 1.53605176524847, 1.54241298337792, 1.54875602626771, 1.55507511477207, 1.56136568270716, 1.56762385269554, 1.57384639125894, 1.58003107047782, 1.58617624104497, 1.59228080478892, 1.59834428373847, 1.60436669894146, 1.61034833855344, 1.61629007677353, 1.62219297364697, 1.62805880524461, 1.63388915306101, 1.63968607693572, 1.6454520901278, 1.65118961694146, 1.65690111890984, 1.66258981922363, 1.66825878629818, 1.67391107830085, 1.67955023403617, 1.68517890218943, 1.69079974948912, 1.69641437838303, 1.70202471010731, 1.70763166851762, 1.71323631309792, 1.71883895859837, 1.72444036202574, 1.73004057600699, 1.73563963802825, 1.74123765381658, 1.74683425086243, 1.75242969974443, 1.75802350982855, 1.7636153048555, 1.76920481084361, 1.77479224587072, 1.78037702574427, 1.78595866540977, 1.79153755443377, 1.79711323880513, 1.80268579842259, 1.80825510826782, 1.81382167385918, 1.81938506363411, 1.82494595688218, 1.83050542748661, 1.83606399854169, 1.84162298907452, 1.84718406429254, 1.852747728261, 1.85831676939687, 1.86389228560237, 1.86947671242034, 1.87507227354385, 1.88068091427872, 1.88630483353614, 1.89194774475376, 1.89761189731552, 1.90329977019013, 1.90901476642492, 1.9147592278953, 1.92053788888662, 1.92635372575522, 1.93220991902594, 1.9381107690584, 1.94405923379357, 1.950060256213, 1.95611727020876, 1.96223429816703, 1.96841298305583, 1.97465537155953, 1.98096419972217, 1.98733776515465, 1.9937782279461, 2.00028311901377, 2.00685433181804, 2.0134889397255, 2.02018492074876, 2.02694410658894, 2.03376179348014, 2.04063710890531, 2.04756937194912, 2.05455497856335, 2.06159336288426, 2.06868261205738, 2.07582086545397, 2.0830046205529, 2.09023590329493, 2.0975097114807, 2.10482473610792, 2.1121819717276, 2.11957875797072, 2.12701462950716, 2.13448719883305, 2.1419988887457, 2.14954551624292, 2.15712513438802, 2.16474387518341, 2.17239548283667, 2.18008437807076, 2.18780727810779, 2.19556666765425, 2.20336574388772, 2.21119878208701, 2.21906941973291, 2.22698205773577, 2.23493480695697, 2.24293300965878, 2.25097155398781, 2.25905654209919, 2.26719527668687, 2.27537932962888, 2.28362096166113, 2.29192080670653, 2.30027952855885, 2.30870270580237, 2.31719669394349, 2.3257634597216, 2.33440500590197, 2.34312913380319, 2.3519446901907, 2.36084323440092, 2.36984638829516, 2.37895267180433, 2.38816669309536, 2.3974934303111, 2.40694575307286, 2.41652262380314, 2.42623847874389, 2.43609278812133, 2.44610183134625, 2.45627494375819, 2.46661267459907, 2.47713485083718, 2.48783244747041, 2.49873779400636, 2.50985379201636, 2.52118247263782, 2.53273926231377, 2.54440279665817, 2.5561160609005, 2.56788748643363, 2.57970669214853, 2.59156182988486, 2.60345523756733, 2.61540659871786, 2.62740531921539, 2.6394392357933, 2.65153103963213, 2.66366987863253, 2.67584337963277, 2.68807792889929, 2.70034186549141, 2.71266375198116, 2.72503238119288, 2.73743498722108, 2.74990532665102, 2.76240829303884, 2.77498078554551, 2.78758597969827, 2.80023652556741, 2.8129478327389, 2.82570852837824, 2.83850546334257, 2.85135577225083, 2.86424685261516, 2.87719856224376, 2.89019917903797, 2.90323515645431, 2.91636738285546, 2.92950965958332, 2.94272471175322, 2.9559595897534, 2.96928421491422, 2.98259927045833, 2.99602405710977, 3.00950399150925, 3.02297516196798, 3.03657236069564, 3.05018336872153, 3.06384496491314, 3.07754284824993, 3.09132098863743, 3.10516899701852, 3.11901029556148, 3.13295621206184, 3.14693022663014, 3.1609861656685, 3.17511338319784, 3.18922267849747, 3.20344899745745, 3.21770382372386, 3.23205424552593, 3.24640249313657, 3.26081762014288, 3.27528649552171, 3.28754788245527, 3.30219388807186, 3.31686139493491, 3.33163320089763, 3.3463878501201, 3.36121668688659, 3.37622764355535, 3.39117157622034, 3.40614554704339, 3.42126474829806, 3.43638635248179, 3.45163103083242, 3.46684254774001, 3.48214729260749, 3.49753420443772, 3.51299048002675, 3.52850132810062, 3.54404996064957, 3.55961726221174, 3.57537260062747, 3.59111910556542, 3.60683088921204, 3.62269107170706, 3.63869537055127, 3.65461051656197, 3.6708713142347, 3.68701037951836, 3.70298531929649, 3.71927379598479, 3.73562181986193, 3.75201168809436, 3.76842341990972, 3.78483438783388, 3.80153176224347, 3.81820445242663, 3.83482245355856, 3.85170026708352, 3.86848651721064, 3.88551555883894, 3.90240684772645, 3.91951572335549, 3.93643008056024, 3.9535269898401, 3.97080506478297, 3.98826204144463, 4.0054168678399, 4.0231911021823, 4.04061583790366, 4.05814659523329, 4.07577071255674, 4.09347341856746, 4.11123788393748, 4.12904466415009, 4.14686985021962, 4.16534497497708, 4.18317065401943, 4.20163867675546, 4.21934796252626, 4.23767970135547, 4.25590810783666, 4.27480156227835, 4.29357949644977, 4.29357949644977]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_cb14_1100 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Campbell & Bozorgnia (2014) GMM, for Vs30 = 1100 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_cb14_1100 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_cb14_1100()
        fi = @SVector [0.001, 0.107151935453737, 0.109647834565003, 0.112201844745437, 0.11481541722802, 0.117489804052062, 0.120226447440993, 0.123026916177356, 0.125892546569393, 0.128824945955999, 0.131825714078132, 0.134896306042729, 0.138038424160509, 0.141253751203674, 0.14454400666565, 0.147910845778061, 0.151356122586865, 0.154881709807325, 0.158489319473994, 0.162181025067334, 0.165958710318836, 0.169824438685114, 0.173780141390804, 0.177828043611803, 0.181970119541417, 0.18620873197305, 0.190546134906447, 0.194984475342449, 0.199526210475698, 0.204173837974327, 0.208929640601348, 0.213796247084416, 0.218776211782296, 0.223872152523143, 0.229086871788025, 0.234422946424962, 0.239883314860486, 0.245470938018794, 0.251188642147164, 0.257039598092891, 0.263026835681957, 0.269153521651332, 0.275422976359149, 0.281838329395954, 0.288403215446539, 0.295120983380726, 0.301995230065901, 0.309029624220641, 0.316227902746474, 0.323593746346755, 0.331131246680177, 0.3388441977443, 0.346736961606689, 0.354813433633056, 0.363078259538029, 0.371535260202758, 0.380189545842012, 0.389045181735103, 0.39810727171514, 0.407380375529606, 0.416869489133838, 0.426579731304667, 0.43651591390959, 0.446683836016586, 0.457088344310272, 0.467735362157079, 0.478630277106537, 0.489778904972026, 0.501187289285747, 0.51286142354831, 0.524807868760027, 0.537032155540723, 0.549541011276866, 0.562341531322902, 0.575440020570062, 0.588843861667092, 0.602560003343154, 0.616595264861423, 0.630957736954371, 0.645654518798835, 0.660693557110258, 0.676083280102802, 0.691831491646336, 0.707945991428218, 0.724436553358675, 0.74131051179263, 0.758577769375232, 0.776247559858663, 0.794328977075606, 0.812830810032516, 0.831764029522309, 0.851138571729408, 0.870963922685582, 0.891251996967617, 0.912011247769634, 0.933254698528794, 0.954992787026339, 0.97723758486003, 1.0000005062702, 1.02329347485744, 1.04713040496121, 1.07152004409775, 1.09647897265996, 1.12201844543814, 1.14815528026766, 1.17489835814305, 1.20226604254383, 1.23027018193224, 1.25892734508558, 1.28825079908837, 1.318259163803, 1.34896505843724, 1.38038665711632, 1.41253792667885, 1.44544207656103, 1.47910913805971, 1.51356162108749, 1.5488172358058, 1.58489465883479, 1.62181367335723, 1.65958733700612, 1.69824539848406, 1.7378031878117, 1.77828083283144, 1.81970452337863, 1.86208794334205, 1.90546520741094, 1.94984557553173, 1.99526746389236, 2.04174258587913, 2.08929688261176, 2.137968038963, 2.18776325462817, 2.23872789932023, 2.29087330073668, 2.34423561131071, 2.39883687869452, 2.45471125310074, 2.51189079709591, 2.57040089862653, 2.63027799720795, 2.69153693925965, 2.75423946975239, 2.81838862731573, 2.88403684826771, 2.9512120467315, 3.01995664775724, 3.09029886559892, 3.16228306310782, 3.23593774690871, 3.31132894099924, 3.3884479777074, 3.46738448436814, 3.54814938834072, 3.63079641398924, 3.71535811383629, 3.80189371410457, 3.89046814362258, 3.98109246261088, 4.07380423161549, 4.16870979107803, 4.26582434364337, 4.3651582267743, 4.46683722178704, 4.57088071215204, 4.67734925332799, 4.78630945056098, 4.89778471638987, 5.01190281582109, 5.12864306805199, 5.24809157322569, 5.37034653309144, 5.49545357878399, 5.62346020566487, 5.75441680055487, 5.88845685362517, 6.02565210012032, 6.16599284660048, 6.30964647562554, 6.45661147569388, 6.60698203384398, 6.76088045132855, 6.91830256679691, 7.07948784544066, 7.24446554472023, 7.41312220760793, 7.58574960014797, 7.76254169346153, 7.94339746512847, 8.12835461278235, 8.31763266541652, 8.51149031718433, 8.70962148345103, 8.91249964207197, 9.12023990175558, 9.33250964785959, 9.55009558905348, 9.77253934629245, 10.0000507473303, 10.2330980382324, 10.4712975342726, 10.7151532131155, 10.9649166549558, 11.2201404474293, 11.4818091737881, 11.7490636915642, 12.0225588043305, 12.3026241711845, 12.5891610107431, 12.8825325867975, 13.1826273392915, 13.489849571919, 13.8040774158165, 14.1257655547166, 14.4547859850637, 14.7909591027162, 15.135516773819, 15.4883873310213, 15.8494511209435, 16.2185340183152, 16.596310619533, 16.9826200621376, 17.3782522441654, 17.7830741509695, 18.196892983888, 18.6206353014046, 19.0554026077809, 19.4985300407455, 19.9523177626254, 20.4180811199692, 20.89272169942, 21.3806209556572, 21.8784563347614, 22.387569371586, 22.9095949660671, 23.442429963151, 23.987749119062, 24.4620357500815, 25.0316841957193, 25.6180740535195, 26.2162255117421, 26.8283164660661, 27.4540559607676, 28.096132775273, 28.7512675224258, 29.4222766176383, 30.1089392418281, 30.8147536270626, 31.5318189879249, 32.2674699378919, 33.0216612714318, 33.7942567772711, 34.5850167285209, 35.388313764356, 36.2195077508899, 37.0620713160324, 37.9266985360148, 38.8133730566014, 39.7219756846267, 40.645001386875, 41.5962694155783, 42.5683179538513, 43.5604570068404, 44.5808076412297, 45.6201979729562, 46.6873914210834, 47.771880369061, 48.8942482643485, 50.0324439699154, 51.1966799577741, 52.399433145376, 53.6281299541173, 54.8672478615497, 56.1594363974676, 57.4592813129032, 58.8139200609086, 60.1905090554239, 61.5868188953173, 63.0203244915951, 64.491094304094, 65.9990752044332, 67.5440188946363, 69.1254709414305, 70.7427414108178, 72.394877541108, 74.080632612254, 75.7984354316832, 77.5788807365932, 79.3904112524522, 81.2304780583983, 83.1342530299609, 85.1041774559473, 87.057646940651, 89.1183937383861, 91.2006398887539, 93.299093611543, 95.5130345351385, 97.7408567124562, 100.035271317829, 1e3]
        Ai = @SVector [1.0, 1.02754725519808, 1.02820826952511, 1.02888599630737, 1.02958091766671, 1.03029346852874, 1.03102414038011, 1.0317734650096, 1.03254191876817, 1.03333005763436, 1.03413844319424, 1.03496760589708, 1.03581815111239, 1.03669068782925, 1.03758584430214, 1.03850424100261, 1.03944656285786, 1.04041351096469, 1.04140575375438, 1.04242407109357, 1.04346920356693, 1.04454195496302, 1.04564310667862, 1.04677353699283, 1.04793407015976, 1.04912565608339, 1.05034923240789, 1.05160572535366, 1.0528961688031, 1.05422162980225, 1.05558313657458, 1.05698183992073, 1.0584188948002, 1.05989550117098, 1.06141294102939, 1.06297246082026, 1.06457544425836, 1.06622331610232, 1.06791749739882, 1.06965954942584, 1.07145103252663, 1.07329359100737, 1.0751889615415, 1.07713887192198, 1.07914525536852, 1.08121001390809, 1.08333518446315, 1.08552288943526, 1.08777533980318, 1.09009480033728, 1.0924837416722, 1.09494462221245, 1.09748016882279, 1.10009305244087, 1.1027862863768, 1.10556272234172, 1.10842568083809, 1.11136784469061, 1.11435355003379, 1.11738141633584, 1.12045031913697, 1.12355927550719, 1.12670731516118, 1.12989390082523, 1.13311829129987, 1.13638018549887, 1.13967921599906, 1.14301524754049, 1.14638830840526, 1.14979850858202, 1.15324621483428, 1.1567315885543, 1.16025523623697, 1.16381793878799, 1.16742036336823, 1.17106355608351, 1.17474853091004, 1.17847636368001, 1.18224854983406, 1.18606641729104, 1.18993161151923, 1.19384598114918, 1.19781129475583, 1.20182909102924, 1.20590103406501, 1.21002786120417, 1.21421064602407, 1.21845006862633, 1.22274654155754, 1.2271002381186, 1.23151168439212, 1.2359810313511, 1.24050823143457, 1.24509370110001, 1.24973701876179, 1.25443874469405, 1.25919884040249, 1.26401760004218, 1.26889519984859, 1.27383192291399, 1.27882851643492, 1.28388468478728, 1.28900180359891, 1.29418021864743, 1.2994214782991, 1.30472560497598, 1.31009459262257, 1.31552915024034, 1.32103083952586, 1.3266005801871, 1.33224028006065, 1.33795072191015, 1.34373376089976, 1.3495905021026, 1.35552379964638, 1.36153418186804, 1.36762450845223, 1.37379687912623, 1.38005366319019, 1.38639754202026, 1.3928301859908, 1.39935624048714, 1.40597807171151, 1.41269903285634, 1.41952366414661, 1.42645419789768, 1.43349703021265, 1.44065490973256, 1.44793524785183, 1.45534128438651, 1.46287749715587, 1.47054851918107, 1.47835265120915, 1.4862927780237, 1.49436524717492, 1.50256914493868, 1.51090030083193, 1.51935419619566, 1.52792774102079, 1.53661503766433, 1.54541127373826, 1.55430795530326, 1.56330300769557, 1.572385566197, 1.58155154577689, 1.59079317030895, 1.60010452227965, 1.6094776474509, 1.61890670396021, 1.62838374933906, 1.63790564281385, 1.64745986727507, 1.65704624667635, 1.66665499794966, 1.67628164844266, 1.6859193933619, 1.69556462637196, 1.70521444996092, 1.71486024791311, 1.72449656312858, 1.73412525344792, 1.7437389062475, 1.75333004291949, 1.76290258170278, 1.77245033628628, 1.78197124553899, 1.7914639265763, 1.80092341488628, 1.8103537955354, 1.81974693738443, 1.82910370605815, 1.83842595995007, 1.84771158226806, 1.85695887787945, 1.86616654981164, 1.87533922635142, 1.88447723531203, 1.89357569237835, 1.90264145845155, 1.91167063861534, 1.92066553222247, 1.92962642827777, 1.93855244152431, 1.94745576840511, 1.95633525932969, 1.96518249341382, 1.97401063388319, 1.982827379544, 1.99162573649249, 2.00040599974566, 2.00917715895918, 2.01794974150186, 2.02670878822112, 2.03547434661141, 2.04425042070715, 2.05297568659385, 2.06177350989524, 2.07060345759516, 2.07945917170465, 2.08835423592739, 2.09726942461957, 2.10621907638622, 2.11520760348828, 2.12421430988326, 2.13326921145113, 2.14233790152115, 2.15143812801247, 2.16057604753109, 2.16974365971957, 2.17894779640365, 2.18818016422479, 2.19744841549616, 2.2067440754367, 2.21607579216373, 2.22543502484972, 2.23481207869526, 2.24423642137433, 2.25370092138661, 2.26319737877306, 2.27271645916396, 2.28227072290856, 2.29185122267207, 2.30147255470604, 2.31112645511759, 2.32080343019524, 2.33052018144713, 2.34029652509159, 2.35006769198692, 2.35987972916044, 2.36975549721751, 2.37962439286674, 2.38957173268132, 2.39952367822879, 2.4095037248115, 2.4195385338672, 2.42958260764245, 2.43966263551656, 2.44826127984448, 2.4584168763621, 2.46866791868729, 2.47892169787595, 2.48921070944454, 2.49952484697339, 2.5099029094798, 2.52028649474119, 2.53071515346962, 2.54117997379675, 2.55172823308834, 2.56223680122082, 2.57280835156284, 2.58343616926067, 2.59411244055014, 2.60482814951695, 2.61550255073071, 2.62633360410579, 2.63710025127131, 2.64793443436663, 2.65882964397415, 2.6697782751069, 2.68068520784913, 2.69170797485315, 2.70275373008123, 2.71380949891863, 2.72495971113114, 2.73609837531697, 2.74731382544185, 2.75849047719981, 2.76983395569256, 2.78111529169552, 2.79243132399273, 2.80389605864023, 2.81538267075703, 2.82674275161086, 2.83836055592965, 2.84982179705231, 2.86153584939425, 2.87321111527154, 2.88482520854258, 2.89651828163281, 2.9082846005574, 2.92011638767014, 2.9320050701364, 2.94394078416375, 2.95591231132987, 2.96790691104069, 2.97991020185413, 2.991905991553, 3.00410033351606, 3.01626982969244, 3.02839336223732, 3.0406957276131, 3.05318299686852, 3.06532876423155, 3.07789607819706, 3.09035278260366, 3.10266558971082, 3.11540789445756, 3.12798668687108, 3.14066648243057, 3.14066648243057]
        new(construct_interpolant(fi, Ai))
    end
end


"""
    struct SiteAmpAlAtikAbrahamson2021_cy14_620 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Chiou & Youngs (2014) GMM, for Vs30 = 620 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_cy14_620 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_cy14_620()
        fi = @SVector [0.001, 0.0999999976652677, 0.102329302559524, 0.104712899705854, 0.107151922989684, 0.109647828995156, 0.112201842696072, 0.11481541089106, 0.117489809378079, 0.120226445776095, 0.123026909458668, 0.125892543532019, 0.128824940148533, 0.131825713266556, 0.134896296983383, 0.138038409311518, 0.141253749305678, 0.144543997655653, 0.147910842278198, 0.151356137537707, 0.15488169337686, 0.158489303595042, 0.162181004512454, 0.165958719179296, 0.169824391893398, 0.173780092333096, 0.177827996145802, 0.181970117416829, 0.186208734415521, 0.190546109357795, 0.194984457628401, 0.199526242691258, 0.204173813286818, 0.208929592034584, 0.213796238584529, 0.2187762493612, 0.223872144370144, 0.229086835045603, 0.23442290250074, 0.239883271573778, 0.245470902647703, 0.25118864038462, 0.257039633852038, 0.263026746068401, 0.269153491549264, 0.27542291591493, 0.281838388993556, 0.288403255925135, 0.295121024058995, 0.301995126127994, 0.309029604841999, 0.316227931529218, 0.323593614975621, 0.331131069452825, 0.338844245205114, 0.346736772309639, 0.354813575838685, 0.363078225718974, 0.371535219004033, 0.380189390728033, 0.389045046146597, 0.398107161691699, 0.407380508152432, 0.416869557404415, 0.426579406652838, 0.436515931430506, 0.446683454307402, 0.457088153096857, 0.467735056430679, 0.478629876540239, 0.489779099866922, 0.50118693836304, 0.512861667967552, 0.524807676772549, 0.537031867432931, 0.549541200445333, 0.56234152115601, 0.575440464532454, 0.588844296122595, 0.602559893922469, 0.616594876327047, 0.630957658073171, 0.645654646063062, 0.66069423786917, 0.676082399828016, 0.691830540350792, 0.707945998165234, 0.724436595494636, 0.741310700488815, 0.758577297086523, 0.776248385474137, 0.794327772623156, 0.812831358119426, 0.831763986545778, 0.851138538780708, 0.870963429537979, 0.89125049813546, 0.912012592563694, 0.933253442052435, 0.954994171529599, 0.977235937307919, 0.999998805375704, 1.02329280366624, 1.04712852187349, 1.07152197527833, 1.09647602264793, 1.12201883715915, 1.14815530337399, 1.17489542276462, 1.20226202750818, 1.23026811928943, 1.25892769214851, 1.2882486488801, 1.31825399709005, 1.34896180039626, 1.38038352538763, 1.41254005943251, 1.44544529540811, 1.47911398130551, 1.51356193372786, 1.54881707745128, 1.58489956887586, 1.62180735282214, 1.65958637551936, 1.69824989899918, 1.73779729406129, 1.77828448946428, 1.81969887518605, 1.86208779112509, 1.9054543333368, 1.94985352787541, 1.99527311144687, 2.04173508683357, 2.0893051336317, 2.13797123362986, 2.18776191655018, 2.23873283609303, 2.29087102929548, 2.34423796842628, 2.39884781988837, 2.45471400144894, 2.511880091243, 2.57039510551677, 2.63027965310721, 2.69155421647587, 2.75423951576535, 2.81839609775087, 2.88405090257416, 2.95123164803845, 3.01996716410099, 3.09028750979791, 3.16227509284513, 3.23591745184899, 3.31130626470979, 3.38842932293713, 3.46738992240694, 3.54811160360915, 3.63076945311162, 3.71535442092829, 3.80192419136001, 3.89046340636454, 3.98111235975458, 4.0737739992345, 4.16868824170493, 4.2657561712032, 4.36515121132422, 4.4668652059137, 4.57088048539599, 4.67739949206702, 4.786295133735, 4.89778717629173, 5.01186630518887, 5.1286521059794, 5.24808226229213, 5.37027721599142, 5.49538226227858, 5.62336817840178, 5.75436927581519, 5.88853864913959, 6.02565307702634, 6.16606017862807, 6.30951109892382, 6.45662098303551, 6.60690524504259, 6.76079098661223, 6.91824567875567, 7.07950157366777, 7.24452892812175, 7.41327778010241, 7.58567623087047, 7.76232417539447, 7.94320265842556, 8.12827310567713, 8.31747401364312, 8.5115727059095, 8.70968476478233, 8.91261544753868, 9.12033265272673, 9.33278033873792, 9.54987524925011, 9.77266347562167, 9.99995236582352, 10.2328495146228, 10.4713310895162, 10.7153467856455, 10.9648168734375, 11.220415214923, 11.481281797322, 11.7489745722971, 12.0226074114367, 12.3030150474414, 12.589154319388, 12.8829520048479, 13.1822172447431, 13.4901993826242, 13.8033298484174, 14.1251356107456, 14.4543495498987, 14.7908540561393, 15.1360143919102, 15.4882383432476, 15.848957724656, 16.2181055509997, 16.5955707472823, 16.9831652499188, 17.3788959801615, 17.782498830651, 18.1982415618891, 18.6215754095712, 19.0545903354785, 19.4998054296033, 19.9517653945402, 20.4186748032945, 20.8916998179073, 21.3798185481755, 21.876592318871, 22.388724887306, 22.9089776595424, 23.4406285116069, 23.9876462125801, 24.4611810638043, 25.0304353881407, 25.6151373396459, 26.2154380805413, 26.8260113521301, 27.4518039604166, 28.0927715913549, 28.7488005082378, 29.4196999330841, 30.1122048289434, 30.8122344526922, 31.5337746712797, 32.269216867667, 33.0178739029947, 33.7970297127673, 34.5798334266531, 35.3939002732706, 36.2198517187626, 37.0674347729588, 37.9250384135495, 38.815527268243, 39.7146910637927, 40.6476598657917, 41.6017669360864, 42.5614794901379, 43.5557410416839, 44.5861329379982, 45.6190677798908, 46.6879684576793, 47.7748219499769, 48.8987423243348, 50.0395561486413, 51.1954464339143, 52.3882302992812, 53.6190884819327, 54.8626899373696, 56.1440795380313, 57.4642183008033, 58.7931299932705, 60.192060004592, 61.5980795163853, 63.0427913269658, 64.4886848417301, 66.0101801162325, 67.5292356025776, 69.1279644465855, 70.7196167323996, 72.3947924900772, 74.057021287325, 75.8062482577471, 77.5922945719624, 79.4139009601905, 81.2694908799235, 83.1571343713345, 85.0745198625965, 87.0926546875546, 89.0642939141243, 91.2194706679196, 93.3207893198619, 95.5324228492403, 97.7678473288137, 100.021978021978, 1e3]
        Ai = @SVector [1.0, 1.43770241338022, 1.45016787169638, 1.46262454089254, 1.47506954446146, 1.4875032820005, 1.49992636965784, 1.51234180697138, 1.5247529223422, 1.53716475354566, 1.54958405816321, 1.56201737581464, 1.57447313258116, 1.58696060323328, 1.59948897271967, 1.61206931166247, 1.62471240114601, 1.63742528130861, 1.65021024760093, 1.66306589337842, 1.67598680368875, 1.6889641838872, 1.70198714824542, 1.71504181008475, 1.72811216588793, 1.7411807603792, 1.75422891589671, 1.76723619345155, 1.78018337692914, 1.79305328990608, 1.80583076079616, 1.8185032630888, 1.83105967940179, 1.84349068085707, 1.85578895086534, 1.86794800786156, 1.87996254635489, 1.89182916026216, 1.90354456326478, 1.91510698529507, 1.92651538016369, 1.93776906215491, 1.94886842373616, 1.95981376147573, 1.97060694489046, 1.98124927169874, 1.99174289867486, 2.00209017936097, 2.01229386115204, 2.02235673002841, 2.0322825729852, 2.04207439107915, 2.05173532871477, 2.06126973721357, 2.07068138943787, 2.07997366611491, 2.08915149661157, 2.09821815137207, 2.107177996807, 2.11603566959446, 2.1247952191208, 2.13346129161745, 2.14203831430708, 2.15053031049951, 2.15894190392058, 2.16727828847451, 2.17554324021112, 2.18374201605683, 2.19187893599667, 2.199958720043, 2.20798664627671, 2.21596612288845, 2.22390264827575, 2.23179796019478, 2.23965456530301, 2.24747416994563, 2.25525698088987, 2.26300369255647, 2.27071359119099, 2.27838589422441, 2.28601973247664, 2.29361432874004, 2.30116749996813, 2.3086778356036, 2.31614246908875, 2.32356100141267, 2.33093092939542, 2.33824981714173, 2.34551537549608, 2.35272554679509, 2.3598793895228, 2.36697264291035, 2.37400578716703, 2.38097495005529, 2.38787931996052, 2.39471648051213, 2.40148541810272, 2.40818570002686, 2.41481417437546, 2.42137350269376, 2.42786038034924, 2.43427730762559, 2.44062404245448, 2.44690066141568, 2.45310879514176, 2.45924683690326, 2.46531959118402, 2.47132617807982, 2.47726715655229, 2.4831459587665, 2.48896386728503, 2.49472238892009, 2.50042193775653, 2.50606585942444, 2.5116565033774, 2.51719510112322, 2.52268444138543, 2.52812616926381, 2.53352210977711, 2.53887424152971, 2.54418639048603, 2.54946111207599, 2.55469777767125, 2.55990258293713, 2.56507702827636, 2.57022091276869, 2.57534127336485, 2.58043642448844, 2.58551213835869, 2.59056876953017, 2.59561273694766, 2.60064266491665, 2.60566114763947, 2.61067538498963, 2.61568429867945, 2.62069108211207, 2.62570155924051, 2.63071474081604, 2.63573686785813, 2.64076965035807, 2.64581473404395, 2.65087659127751, 2.6559599807492, 2.66106717721275, 2.66620056913522, 2.6713624883885, 2.67655862119629, 2.68179180173664, 2.68706503247137, 2.69238140763953, 2.69774416968228, 2.70316051272804, 2.70863043864547, 2.71416182660125, 2.71975477246343, 2.72541777755935, 2.73114690776991, 2.73695590437774, 2.74284555208644, 2.74882134784057, 2.75488382444996, 2.76104426912801, 2.76729788279163, 2.7736627502061, 2.78013423665675, 2.78672602937371, 2.79343983347924, 2.80027692924534, 2.80725315077232, 2.814362847559, 2.82162331360779, 2.82903694748673, 2.83661476441152, 2.8443206041498, 2.8520511877403, 2.8598037746113, 2.86757243990086, 2.875361246825, 2.88317505759392, 2.89099690156617, 2.89884252206407, 2.90669402806585, 2.91458089482491, 2.92247297709132, 2.93038865266971, 2.93832205666209, 2.94628067082922, 2.95425865017668, 2.9622495348731, 2.97024603444418, 2.97827180113265, 2.98632157144484, 2.99438928056263, 3.00246817559146, 3.01058658006503, 3.0187032248018, 3.026847073473, 3.03501243241576, 3.04319282444565, 3.05138099124304, 3.05961193178662, 3.06783730096053, 3.07609307589279, 3.08437380669606, 3.09267333573152, 3.10098467953501, 3.10932597146677, 3.11766492555021, 3.12604695408217, 3.13443972680896, 3.14286435755985, 3.15128526682163, 3.15975469356089, 3.16820522174391, 3.176723970201, 3.18520799445835, 3.19374858930422, 3.20230709657814, 3.2108761883538, 3.21948596929976, 3.22809231538926, 3.23672600296876, 3.24538073379847, 3.25404942554169, 3.26276879153558, 3.2714892521925, 3.28020129743411, 3.28899195109764, 3.29776022041696, 3.30654556003318, 3.31539381860346, 3.3241924704965, 3.33309635042727, 3.34193250145537, 3.35086420237266, 3.35976851638599, 3.36876064179229, 3.37770877149997, 3.38666591577686, 3.39569365553722, 3.40336128011046, 3.41240444748267, 3.42150263892667, 3.43065244885735, 3.43976835149571, 3.4489200680504, 3.4581017634771, 3.46730683156448, 3.47652774775042, 3.48585109898529, 3.49508310600139, 3.5044040094006, 3.51371021423701, 3.52298951241011, 3.53244940197633, 3.5417593689859, 3.55124314831317, 3.56066891069682, 3.57014379913305, 3.57953452542909, 3.58908587962355, 3.59853308319482, 3.60813502113749, 3.61775428403747, 3.62723216085909, 3.63685006074365, 3.64661438492979, 3.65620309560786, 3.66592287374757, 3.67560418251516, 3.68541142885049, 3.69516310864239, 3.70484159030479, 3.71462470355701, 3.7245142324531, 3.73430225802669, 3.74418162900315, 3.7541524596334, 3.76398469248638, 3.77412429347026, 3.78410841347155, 3.79415840059195, 3.80401131424911, 3.81416782825724, 3.82410151631852, 3.83434326044182, 3.84433203673702, 3.85463092634442, 3.86464222983466, 3.87496297029135, 3.88528746895873, 3.89560403281376, 3.90589970622295, 3.91616032332114, 3.92637033463232, 3.93689861761273, 3.94697505070503, 3.95776590239446, 3.96807466174167, 3.97870464642664, 3.98923199048823, 3.99963202873836, 3.99963202873836]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_cy14_760 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Chiou & Youngs (2014) GMM, for Vs30 = 760 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_cy14_760 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_cy14_760()
        fi = @SVector [0.001, 0.100000005278119, 0.102329305685959, 0.104712901088038, 0.10715191734136, 0.1096478161355, 0.112201844312325, 0.114815399493054, 0.117489811269133, 0.120226426884127, 0.123026912673429, 0.125892543180806, 0.128824943888933, 0.131825701931171, 0.134896292542321, 0.138038419685919, 0.141253726031804, 0.144543991682274, 0.147910841016824, 0.15135612531762, 0.154881699105396, 0.158489298275321, 0.162180994731809, 0.165958688177051, 0.169824406249264, 0.173780124732544, 0.177827987234069, 0.181970104503309, 0.18620874216892, 0.190546088276735, 0.194984413397179, 0.199526234462532, 0.204173818378493, 0.208929629962364, 0.213796203501538, 0.218776216272469, 0.223872092751581, 0.229086773295876, 0.234422913440041, 0.239883346374602, 0.245470959191782, 0.251188647299664, 0.257039583965111, 0.263026839625322, 0.269153386068061, 0.275422892536035, 0.281838357206786, 0.288403136014915, 0.29512088643951, 0.301995147118928, 0.309029636270837, 0.316227726764557, 0.323593743035287, 0.331131074274485, 0.338844138033324, 0.346736738117816, 0.354813374379999, 0.363078069427477, 0.371535318437207, 0.38018929961147, 0.389044995694541, 0.398107060813636, 0.407380379735427, 0.416869369537323, 0.426579727361764, 0.436515887692042, 0.446683606986212, 0.457088019319652, 0.46773532542682, 0.478629981874554, 0.489778948828998, 0.501187304961605, 0.512861186273828, 0.524807402402299, 0.537031596210482, 0.549541299892976, 0.562341327015613, 0.575439822942147, 0.588843516357273, 0.602559763585899, 0.61659488244124, 0.630957589108888, 0.645654602589703, 0.660693172670039, 0.676083352446191, 0.69183061540227, 0.707945609900711, 0.724435639005472, 0.741310319666438, 0.758577791401586, 0.776247032532349, 0.794327972885161, 0.812829888313304, 0.83176450189098, 0.85113745411528, 0.870963690655165, 0.891251760400126, 0.912010408577856, 0.933255245295474, 0.954991724274509, 0.977236931585538, 0.99999931915226, 1.0232926583713, 1.04712871370939, 1.07151972559382, 1.09647839861261, 1.12201798920987, 1.14815239015486, 1.17489619113164, 1.20226482282661, 1.230270509361, 1.25892568257472, 1.28825221490343, 1.31825970420816, 1.34896234511787, 1.38038569480746, 1.4125355875376, 1.44544035634817, 1.4791067005516, 1.51355991769356, 1.54881419338307, 1.58489135570109, 1.62180741098713, 1.65958724224312, 1.69824098493074, 1.73780538614063, 1.77828384029801, 1.81969768325487, 1.86209132642581, 1.90545951786295, 1.94984000723927, 1.99526323502638, 2.04173666608573, 2.08929373051144, 2.13795720949869, 2.18776679563686, 2.23871864940605, 2.29087274897834, 2.34422665758295, 2.398830455509, 2.45470143491059, 2.51187774934974, 2.57040187591233, 2.63027473614656, 2.69154228983537, 2.75423051595883, 2.81839385215467, 2.88403692114349, 2.95122086420034, 3.01995215562392, 3.09030065678915, 3.16227642295472, 3.23592346631478, 3.31132891490246, 3.3884299172534, 3.46735946910046, 3.54813821074709, 3.63078533725135, 3.71536846377799, 3.80191135853523, 3.89043708383884, 3.98108381905827, 4.07382969500055, 4.16870567544569, 4.26581070164795, 4.36518697265844, 4.46680411703705, 4.57085377834242, 4.67739610111622, 4.78632242414544, 4.89776678860413, 5.0118843229469, 5.12865332406043, 5.24803935339088, 5.37032511094351, 5.49537847413284, 5.62339713074269, 5.75435457691296, 5.88847905118778, 6.02561155450687, 6.16599641750948, 6.30960714497232, 6.45659960120723, 6.60689861529496, 6.76084973072587, 6.91824357308313, 7.07938541613807, 7.24442301919855, 7.41307142632331, 7.58568641738688, 7.76244066564601, 7.94326569146627, 8.12834448596361, 8.31759182767968, 8.51151566867657, 8.70973318814878, 8.91245786665494, 9.12030564789694, 9.33246351716575, 9.54996668260048, 9.77234156401459, 9.99990153906584, 10.2330151684908, 10.4711008559889, 10.7150364469821, 10.964750490257, 11.2201347329491, 11.4816670597186, 11.7492496241653, 12.022742813086, 12.3026981000034, 12.5889920759082, 12.8822769465087, 13.182459265256, 13.489400932834, 13.8038791736706, 14.1258004265005, 14.4539459356652, 14.7913578984967, 15.1357324856635, 15.4880193842308, 15.8493619845298, 16.2183086824161, 16.5960326767835, 16.9823769767776, 17.3787778053556, 17.7834601578259, 18.1979135417807, 18.6200418512582, 19.0554824815177, 19.49807748989, 19.9517012657342, 20.4186610376656, 20.8940733113534, 21.379997361205, 21.8789412825561, 22.387937087537, 22.9096392147486, 23.4439269263859, 23.9870856553391, 24.46058662679, 25.0340588613218, 25.6152039490899, 26.2158596298214, 26.8275446219964, 27.4542925051998, 28.0959151088508, 28.7521309339848, 29.4225501552548, 30.1066616578723, 30.8101318295223, 31.5331607803149, 32.2688473500071, 33.0235676906087, 33.7894041350745, 34.5814827413102, 35.3920282683592, 36.2114034095699, 37.0570384215831, 37.9296731960298, 38.808356830246, 39.7246511532133, 40.6446342457211, 41.5904771517877, 42.5624316552899, 43.5606664842275, 44.58525356566, 45.6203356517073, 46.679926731282, 47.7811828082431, 48.889095202211, 50.0392749799897, 51.1923784631563, 52.3876989427786, 53.6271327545174, 54.8639723203735, 56.1438945185322, 57.4685744162854, 58.8111011775592, 60.1686734628619, 61.6017896398734, 63.0158092367494, 64.5080917430415, 66.010239337733, 67.5570266517399, 69.1074610525002, 70.7437395379651, 72.3800126937281, 74.0584033514694, 75.8305646487854, 77.5949512656221, 79.4004309784528, 81.2461188644527, 83.1308139099541, 85.0529665055861, 87.0822300450569, 89.0763954752011, 91.1807407221922, 93.3202913166248, 95.4916631049442, 97.6908631563263, 100.012267904509, 1e3]
        Ai = @SVector [1.0, 1.26474365754693, 1.27187168763244, 1.27901461204394, 1.2861668468288, 1.29332769860997, 1.30049545713136, 1.30766852765824, 1.3148448358383, 1.32202240367609, 1.32920111176275, 1.33638102751286, 1.34356422475535, 1.35075365610266, 1.35795288790118, 1.36516683588672, 1.37240078818257, 1.37965956984633, 1.38694651866824, 1.3942644665929, 1.40161534480099, 1.40900004959024, 1.41641948596057, 1.42387361692424, 1.43136209798135, 1.43888401061133, 1.44643774756037, 1.45402034276764, 1.46162798960857, 1.46925573360015, 1.47689789847254, 1.48454848588452, 1.49220043010574, 1.49984643882526, 1.50747891120157, 1.51509009761108, 1.52267156234887, 1.53021542332592, 1.53771329159611, 1.54515767577195, 1.5525423017734, 1.55986193063354, 1.56711260297083, 1.57429107685139, 1.58139470109949, 1.58842230933339, 1.59537254872675, 1.60224497526539, 1.6090399193016, 1.61575793863017, 1.62240013364471, 1.62896756722738, 1.63546241542315, 1.64188624895108, 1.64824169962538, 1.65453105215902, 1.660757253143, 1.66692298962119, 1.67303139815318, 1.67908545861599, 1.68508877287354, 1.69104474095905, 1.69695691445104, 1.70282847890391, 1.70866337029487, 1.71446466460501, 1.72023619436562, 1.72598137429174, 1.73170412937894, 1.73740741092509, 1.74309547823134, 1.74877153446106, 1.7544392973187, 1.76010278844154, 1.7657654338474, 1.77143153644529, 1.77710410272141, 1.7827876649385, 1.78848607318854, 1.79420347226643, 1.79994358407784, 1.80571108575488, 1.81150988535825, 1.81734412132426, 1.82321904347219, 1.82913817252675, 1.83510670026067, 1.84112756121076, 1.84720347424249, 1.85333562714735, 1.85952470296974, 1.8657708979849, 1.87207351949043, 1.87843213317722, 1.8848436987637, 1.89130781117208, 1.89782159094484, 1.90438189213204, 1.91098741512956, 1.91763345905683, 1.92431870117879, 1.93103904302077, 1.9377917893536, 1.94457361968632, 1.95138121657727, 1.95821137709387, 1.96506101732005, 1.97192719683423, 1.97880718423323, 1.98569844861012, 1.99259770558709, 1.99950182429154, 2.00641002878188, 2.01331868946611, 2.02022543975041, 2.02713050388327, 2.03402982723964, 2.04092437374658, 2.04781056500155, 2.05468885904031, 2.0615574700588, 2.06841626488153, 2.0752640907423, 2.08210150758018, 2.08892649642953, 2.09574183898203, 2.10254459906843, 2.10933506179724, 2.11611724811337, 2.12288728136106, 2.12964821625687, 2.13640189502178, 2.14314680483325, 2.14988534511825, 2.15661842855298, 2.16334930355606, 2.17007541648867, 2.17680275350116, 2.18352920748937, 2.19025940643645, 2.19699395722506, 2.2037360384332, 2.21048929651311, 2.21725267854449, 2.22403035334881, 2.23082429847101, 2.23763960582685, 2.24447608988636, 2.25133959947153, 2.25823040581599, 2.26515517696813, 2.27211465284156, 2.279113030692, 2.28615863788022, 2.29324589179053, 2.30038750095143, 2.30758593730311, 2.31484372550238, 2.32216775190769, 2.32956116012962, 2.33702714372695, 2.34457871599795, 2.35221548594714, 2.35994167884065, 2.36776709937289, 2.37569708603435, 2.38373136578869, 2.39188740715789, 2.4001723655651, 2.40858055742088, 2.41712532444938, 2.42582176653978, 2.43467176969305, 2.44367644631041, 2.45286124704223, 2.46222059889919, 2.47177408512471, 2.48152482557902, 2.49149534670758, 2.50167973724668, 2.5121025424396, 2.52276859362371, 2.53368621015369, 2.54471244064013, 2.55579887509383, 2.56691556730856, 2.57807828112741, 2.58929124755218, 2.60052938345122, 2.61181075197994, 2.6231406506567, 2.63450877844716, 2.64592079024619, 2.65736566297975, 2.66886801235028, 2.68039925680379, 2.6919660621504, 2.70359735310195, 2.71524178942724, 2.72695021810889, 2.73869095536778, 2.75047462230107, 2.76231402600025, 2.77417366427405, 2.78609125336844, 2.79805689620922, 2.81005922472766, 2.82211450958847, 2.83421194016232, 2.84633914825058, 2.85851444304052, 2.87072623771478, 2.88299593842877, 2.89531316315953, 2.90766594630464, 2.92007909343311, 2.93254220452969, 2.94500235859685, 2.95756836203094, 2.97014763145451, 2.98276880579325, 2.99546605253435, 3.00818177435952, 3.02095009321979, 3.03375911466721, 3.04664949155654, 3.05955682058631, 3.07252220143771, 3.08547407825541, 3.09857806889149, 3.11164189991262, 3.1247740454825, 3.13803326445637, 3.15127392356254, 3.16454764394741, 3.17791564782097, 3.19129151012199, 3.20473834661461, 3.21824566667958, 3.23171364822652, 3.24322176393691, 3.25693690766748, 3.27057303493195, 3.28439587776679, 3.2982023981508, 3.3120769920673, 3.32600819520374, 3.33998266349085, 3.35398523152793, 3.36799876948903, 3.38213175244099, 3.39637913151413, 3.41059795760726, 3.42490457070842, 3.43914277447917, 3.45358581167583, 3.46808249458969, 3.4824554766011, 3.49700354471659, 3.51172864029507, 3.52627127838015, 3.54114568729358, 3.55579375948015, 3.57056350893963, 3.58544956703106, 3.60044533617299, 3.61554285093282, 3.63050282206819, 3.64552230305604, 3.66083358332543, 3.67594250063318, 3.69132730927766, 3.70645512356833, 3.72183502705042, 3.73747813266146, 3.75278864636778, 3.76832751756716, 3.78410258957844, 3.79978518957619, 3.81533904720756, 3.83144482482593, 3.84703224899097, 3.86316870335808, 3.87910227339303, 3.89519546271496, 3.91101665356257, 3.92739373576442, 3.94345781904698, 3.95961844868298, 3.97635732852111, 3.99270630339808, 4.00911573850308, 4.02556924273202, 4.04204836767728, 4.05853233081845, 4.07560361892911, 4.09205960420983, 4.10909253284836, 4.12608173443567, 4.14299493817811, 4.15979664012907, 4.17719144318364, 4.17719144318364]
        new(construct_interpolant(fi, Ai))
    end
end

"""
    struct SiteAmpAlAtikAbrahamson2021_cy14_1100 <: SiteAmplification

Implementation of the Al Atik & Abrahamson (2021) amplification function for the Chiou & Youngs (2014) GMM, for Vs30 = 1100 m/s
"""
struct SiteAmpAlAtikAbrahamson2021_cy14_1100 <: SiteAmplification
    amplification::Function

    function SiteAmpAlAtikAbrahamson2021_cy14_1100()
        fi = @SVector [0.001, 0.0977010025977648, 0.100000002147813, 0.102329309655048, 0.104712908884942, 0.107151921599752, 0.109647820058097, 0.112201843942563, 0.114815401321503, 0.117489803171727, 0.120226429999859, 0.123026915212088, 0.125892545558633, 0.128824944897604, 0.131825712969856, 0.134896304882222, 0.138038422945309, 0.141253749931204, 0.144544005333209, 0.147910844382825, 0.151356121125874, 0.154881708277479, 0.158489317872049, 0.162181023389891, 0.165958708562337, 0.169824403885452, 0.173780104951092, 0.177828005454721, 0.181970117429641, 0.18620872976175, 0.190546132590931, 0.194984429467593, 0.199526207936787, 0.204173835315759, 0.208929637817485, 0.213796244169354, 0.218776208729851, 0.223872149326841, 0.229086808463051, 0.234422942920286, 0.239883311190641, 0.245470934175994, 0.25118863812326, 0.257039593879345, 0.263026831269832, 0.26915351703127, 0.275422971521348, 0.281838324330158, 0.288403210141998, 0.295120977826189, 0.301995224249587, 0.309029618130213, 0.31622789636901, 0.323593739668733, 0.331131114375592, 0.338844190421997, 0.346736816537615, 0.354813425604309, 0.363078100472879, 0.371535251399413, 0.380189508791034, 0.389045231112729, 0.398107137135727, 0.407380370581322, 0.416869580502438, 0.426579602266325, 0.436515915885663, 0.446683660333047, 0.457088251100418, 0.467735265470384, 0.478630196128666, 0.489779026173635, 0.501187390878715, 0.512861330755075, 0.524807658522541, 0.537032041448618, 0.549541127929602, 0.56234162744386, 0.575440262356433, 0.588843710293594, 0.602559835951633, 0.616595216384171, 0.630957615277879, 0.645654729513682, 0.660693600957462, 0.676083165254278, 0.691831208109039, 0.707945826928056, 0.724436133611112, 0.74131042217303, 0.758578038308524, 0.77624736932336, 0.794328682342172, 0.812830431366386, 0.831763953983852, 0.851138672022375, 0.87096442965993, 0.891251525828283, 0.91201081083974, 0.933254870536159, 0.954992557324465, 0.977237796174617, 1.00000050289132, 1.02329361857925, 1.04712955845278, 1.07151999189293, 1.09647874351814, 1.12201896666043, 1.14815462914034, 1.17489878135328, 1.20226520272396, 1.23027062883789, 1.25892726545553, 1.2882504677093, 1.31825698358287, 1.34896511219286, 1.38038441048487, 1.41253793942796, 1.44543969181007, 1.47911015332597, 1.51356193891812, 1.54881767314568, 1.58489488713915, 1.621811871956, 1.65958777981605, 1.69824683685932, 1.73780222064615, 1.77828027313385, 1.81970479554162, 1.86209113089599, 1.90546494045257, 1.94984823759762, 1.99526402188307, 2.04174270078569, 2.08929854854814, 2.13796117486418, 2.18776490616744, 2.23872629924922, 2.2908704807379, 2.3442329469835, 2.39883512074943, 2.45470793115014, 2.5118939542785, 2.57039815360423, 2.63027562282402, 2.69154135747925, 2.75423171284814, 2.8183845786919, 2.88403946249489, 2.95120968480996, 3.01995159639335, 3.09030429791269, 3.16227816307602, 3.23594997794318, 3.31131159033704, 3.38844532186045, 3.46738106905059, 3.54814747811791, 3.63079559295023, 3.71535519042189, 3.80190823989626, 3.89046320082663, 3.98108192520036, 4.07380120055439, 4.16869189434466, 4.26579788322738, 4.36516522519641, 4.4668426288625, 4.57088216476806, 4.67734021450466, 4.78632839646284, 4.89777907530505, 5.01190181635595, 5.12862260920556, 5.24807767416142, 5.37030248071806, 5.49539859737328, 5.6234130037025, 5.75439508692842, 5.88847553534566, 6.02564127802564, 6.16595510961186, 6.30957907152021, 6.45660585885845, 6.60693292036847, 6.76086350668524, 6.91829867839757, 7.07947441276088, 7.24441264482972, 7.41313009899397, 7.58578237749003, 7.76255094245573, 7.9433251675949, 8.1282996056464, 8.3177033323344, 8.51142533453252, 8.7097194742949, 8.91246199817186, 9.12016156219638, 9.332696243291, 9.55003805171751, 9.77239450509741, 10.0000029907508, 10.2330553726513, 10.4714586173544, 10.7154077869287, 10.9647806593532, 11.2201385083976, 11.4817667385852, 11.749182692662, 12.0226308508435, 12.3028451426025, 12.5892753879456, 12.8827326835533, 13.1825992921692, 13.4897749635231, 13.8041598088759, 14.1256063708026, 14.4545696058549, 14.7908914452525, 15.1358192657506, 15.4885347993343, 15.8488225194303, 16.2181192725222, 16.5963419405375, 16.9823973303689, 17.3779567228685, 17.7829075841531, 18.1970772186461, 18.6214126043814, 19.0545338392523, 19.4986930161533, 19.9524286446679, 20.416864999908, 20.8933470739312, 21.3801959413476, 21.877050819294, 22.3871039554581, 22.908439141657, 23.4427335942745, 23.9877078447875, 24.4615823340943, 25.031202431116, 25.6177047771876, 26.2161483984282, 26.8259478798891, 27.4552051755687, 28.0951333985226, 28.751315395343, 29.4236385219517, 30.1082477908273, 30.811958621233, 31.5308613988034, 32.2686893997017, 33.0208884169828, 33.7914032028601, 34.579975130502, 35.391548861299, 36.2152106058876, 37.0613062814607, 37.9299347343757, 38.8144960833807, 39.7208188221605, 40.6486334289266, 41.5975541496905, 42.5669911664598, 43.5648564777589, 44.5824470987695, 45.6186722516438, 46.6823777382087, 47.7735736024154, 48.8921454797594, 50.0378325855946, 51.1975836707781, 52.395445399495, 53.6185421737629, 54.8658219542705, 56.1516372230445, 57.4603259933358, 58.8075338769039, 60.175447350898, 61.5810960994084, 63.024723286493, 64.5063997612122, 66.0032231305981, 67.5354572618034, 69.1279139189578, 70.7299653329548, 72.3931067306596, 74.0903777983319, 75.8201354700951, 77.5803564100812, 79.4038019330799, 81.2559529418577, 83.1336398315252, 85.0745163971121, 87.0807659823219, 89.1084652652259, 91.2011614865688, 93.3092685329326, 95.4805967926651, 97.7163690175498, 100.018001618123, 1e3]
        Ai = @SVector [1.0, 1.02292145226887, 1.02347193269187, 1.02403057751128, 1.02460319134443, 1.02519011266872, 1.02579176831187, 1.02640853342845, 1.02704082826441, 1.02768905430581, 1.02835363474967, 1.02903505952671, 1.02973373980831, 1.03045017460176, 1.0311848633099, 1.03193827633036, 1.0327109513952, 1.03350342871595, 1.03431626515321, 1.0351500096106, 1.03600526856439, 1.03688266214622, 1.03778277978816, 1.0387063108895, 1.03965390800433, 1.04062627142124, 1.04162410427155, 1.04264817195258, 1.04369920691201, 1.04477802541463, 1.04588545877383, 1.04702230584016, 1.04818950553136, 1.04938795776172, 1.05061856972311, 1.05188234591929, 1.05318029207605, 1.05451345188998, 1.05588292447437, 1.05728983693677, 1.05873534230198, 1.06022069021328, 1.06174710746976, 1.06331594325248, 1.06492854161542, 1.06658631765947, 1.06829076469872, 1.07004336284863, 1.0718457712776, 1.07369961506345, 1.07560663398471, 1.07756863757093, 1.07958750732489, 1.08166516497738, 1.08380367238586, 1.08600521377592, 1.08827195000029, 1.09060628203965, 1.0930105879726, 1.09548742323904, 1.09803949700572, 1.10066953312697, 1.10338046279428, 1.10617545421057, 1.10905764392622, 1.11203034763532, 1.11509721225581, 1.11825553420698, 1.12144917852688, 1.1246796539187, 1.12794829299106, 1.13125624742406, 1.13460429000094, 1.13799306931939, 1.14142322536332, 1.14489489015684, 1.14840831096586, 1.15196361033023, 1.15556079375142, 1.15919975704052, 1.16288063682518, 1.16660315861036, 1.17036730769621, 1.17417302685345, 1.17802006078517, 1.18190862619046, 1.18583863984576, 1.18981011780842, 1.19382332702879, 1.19787836695458, 1.20197562199144, 1.20611528409883, 1.21029802809752, 1.21452416846484, 1.21879471752033, 1.22311033084606, 1.22747180290555, 1.23188011703117, 1.23633643956717, 1.24084239220589, 1.24539887918669, 1.25000795860436, 1.25467094418524, 1.2593898676351, 1.26416675226572, 1.2690035743003, 1.27390284995903, 1.27886707264258, 1.28389900631067, 1.28900137786973, 1.29417717088048, 1.29942997972775, 1.30476222314791, 1.3101765352044, 1.31567550654443, 1.32126173528755, 1.32693599996809, 1.33270134122456, 1.33855901638881, 1.34451134371996, 1.3505591519097, 1.35670499399946, 1.36295050038664, 1.36929744696008, 1.37574778516408, 1.38230438623046, 1.388968313675, 1.39574296878562, 1.40263144734258, 1.40963555128996, 1.41675897819501, 1.42400498055988, 1.43137720353972, 1.43888069172586, 1.44651763482068, 1.45429126519843, 1.46220404573045, 1.47025422831336, 1.4784401930991, 1.48676085496561, 1.49521194341356, 1.50378978556612, 1.51249174592764, 1.52130875754207, 1.53023878441403, 1.5392732835319, 1.54840646886376, 1.55763243232209, 1.56694520621564, 1.57633446807761, 1.58579484589713, 1.5953215555526, 1.60490419160944, 1.61454123113567, 1.6242199810549, 1.63393955229688, 1.64369237447499, 1.65347088883277, 1.66327040535463, 1.67308386455574, 1.68291038791816, 1.69274086388146, 1.70257239762523, 1.71239953852971, 1.72222060085856, 1.73203128225791, 1.74182768716387, 1.75160641484959, 1.76136461193462, 1.7710999990701, 1.78081437519508, 1.79049296371695, 1.80015003736033, 1.80977303709577, 1.81936727071378, 1.82892990714556, 1.83846347170023, 1.84796639243239, 1.85743750459596, 1.86688165331811, 1.87629358252266, 1.8856736527025, 1.8950288840274, 1.90436165066947, 1.91366219165397, 1.92294620623536, 1.93220485121666, 1.94144937133627, 1.95067852479334, 1.95989108471327, 1.96909361501791, 1.97829392522576, 1.9874846673211, 1.99667446850001, 2.00587343176446, 2.01507515603444, 2.02429093618278, 2.03351451411184, 2.0427685885531, 2.05200835048918, 2.06129846975702, 2.07063233834377, 2.08000295297832, 2.08941323937857, 2.09885457878688, 2.10832985289235, 2.11782956098308, 2.12737020834432, 2.13695728579534, 2.14656816953363, 2.15620690857747, 2.16589428783657, 2.17560626370137, 2.18536526516532, 2.19514576543929, 2.20497210102333, 2.2148358498885, 2.22472736252484, 2.23465562018838, 2.24461085240881, 2.25462463207177, 2.26466785090401, 2.27472952825745, 2.28484459422524, 2.29500521882397, 2.30517703409137, 2.31539895840579, 2.32566256141126, 2.3359582202847, 2.34630404776055, 2.35666125692194, 2.36707841037313, 2.37751582151233, 2.38799409394066, 2.39853782035382, 2.4091042877289, 2.41968080309917, 2.43032975604236, 2.44100536528213, 2.45173644910413, 2.46247192861055, 2.47162153106358, 2.48244396248787, 2.49337663294825, 2.50431761351121, 2.5152516264796, 2.52631760041526, 2.53735514381143, 2.54845541134747, 2.5596102315799, 2.5707504876661, 2.5819814097096, 2.59323437974958, 2.60456179747344, 2.61588810012201, 2.6272672323506, 2.63868932517327, 2.65021919295664, 2.66169639382403, 2.67325977972402, 2.68490368042331, 2.69653388068794, 2.70822161597118, 2.71995708946518, 2.73172820368481, 2.74352318663082, 2.7554315670374, 2.76734304435521, 2.77924007834796, 2.79121828669927, 2.80327079986881, 2.81538948486291, 2.82756495832855, 2.83965389962159, 2.85190084056504, 2.86416685159584, 2.87643600596561, 2.88884248273541, 2.90122863794132, 2.91373609924084, 2.92619338563946, 2.93874979637281, 2.95139980678776, 2.96413759574585, 2.97676045753032, 2.98943468223343, 3.00235625109584, 3.0151082390744, 3.02809427260794, 3.04109550561551, 3.05409394159766, 3.06706942999362, 3.08025557152477, 3.09339547933017, 3.10646287098568, 3.11971287329116, 3.13315030041192, 3.14647480120328, 3.15996638094659, 3.17329994536433, 3.18677262243827, 3.20038298587724, 3.21412318639176, 3.21412318639176]
        new(construct_interpolant(fi, Ai))
    end
end

